#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 16:32:10 2024

@author: ncro8394
"""

import re
from itertools import product
from numpy import (asarray, full, nan, ndarray, sum)
from os import listdir, mkdir, path, walk
from pandas import concat, DataFrame, read_csv
from pathlib import Path
from wonambi import Dataset
from wonambi.attr import Annotations
from wonambi.trans import (fetch, get_times)
from wonambi.trans.analyze import event_params, export_event_params
from ..utils.logs import create_logger, create_logger_empty
from ..utils.load import (load_channels, load_adap_bands, rename_channels, 
                          read_manual_peaks, read_tracking_sheet,
                          reverse_chan_lookup)

class FISH:
    
    '''
        Functions for Information Saving and Hypothesis testing (FISH)
    
        
    '''
    
    
    def __init__(self, rootpath, rec_dir, xml_dir, out_dir, chan, ref_chan, 
                 grp_name, stage, rater = None,
                 subs = 'all', sessions = 'all', tracking = None):
        
        self.rootpath = rootpath
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.chan = chan
        self.ref_chan = ref_chan
        self.grp_name = grp_name
        self.stage = stage
        self.rater = rater
        
        self.subs = subs
        self.sessions = sessions
        if tracking == None:
            tracking = {}
        self.tracking = tracking

    def line(self, keyword = None, evt_name = None, cat = (0,0,0,0), 
             segs = None,  cycle_idx = None, frequency = None, 
             adap_bands = 'Fixed', peaks = None,  adap_bw = 4, 
             param_keys = 'all', epoch_dur = 30, Ngo = False, 
             filetype = '.edf',
             logger = create_logger('Export params')):
                           
            '''
                Listing Individual aNnotated Events (LINE)
            
                Extracts event parameters per participant and session.
                
                segs: to extract parameters between certain markers, 
                      these need to be defined in the Annotations file first. 
                      Format should be a list of tuples, with both tags named
                        e.g. [('N2_ON','N2_OFF'), ('N3_ON','N3_OFF')]
                        
                concat = (cycles, stages, discontinuous, evttypes) 
            '''
            ### 1.b. Format event names
            if evt_name == None:
                evt_name = ['spindle']
            if type(evt_name) is not list:
                evt_name = [evt_name]
            
            ### 0.a Set up logging
            flag = 0
            
            logger.info('')
            logger.debug(r""" Exporting event parameters...
                         
                       
                                        /^. 
                                       /   .
                                    o /     .
                                   (√        .
                                   |          .
                        _________/\____"      .
                       |  |     |    |         .
                       |  |     |    |          .  
                       |  |     |    |           .
                       !^~!~^~^~!~^~^!~^~^~^~^~^~^.~^~^~^~^~^~^~^~^~^~^~^~^~ 
                                                  .
                                                  .          œ«   
                                                  ¿   
                               ∞«                    <>≤   
                                                             »<> 
           
                             /\____^__                  ~~~•
                             | __   v  \_____
                             \/  \_____/
                             
                                                 ___:____     |"\/"|
                                               ,'        `.    \  /
                                              |  O   v    \___/  |
                                              \_________________/   
               
                Listing Individual aNnotated Events 
                (L.I.N.E) 
                                                        """,)
            
            ### 1.a. Set up organisation of export
            if cat[0] + cat[1] == 2:
                model = 'whole_night'
                logger.debug('Exporting parameters for the whole night.')
            elif cat[0] + cat[1] == 0:
                model = 'stage*cycle'
                logger.debug('Exporting parameters per stage and cycle separately.')
            elif cat[0] == 0:
                model = 'per_cycle'
                logger.debug('Exporting parameters per cycle separately.')
            elif cat[1] == 0:
                model = 'per_stage'  
                logger.debug('Exporting parameters per stage separately.')
            if 'cycle' in model and cycle_idx == None:
                logger.info('')
                logger.critical("To export cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
                return
            cat = tuple((cat[0],cat[1],0,0)) # force non-concatenation of discontinuous & events
            
    
            ### 1.c. Set default paramaters to export (if not pre-set)
            if param_keys is None:
                params = {k: 0 for k in ['dur', 'minamp', 'maxamp', 'ptp', 'rms',
                                          'power', 'peakpf', 'energy', 'peakef']}
            elif param_keys == 'all':
                params = {k: 1 for k in ['dur', 'minamp', 'maxamp', 'ptp', 'rms',
                                          'power', 'peakpf', 'energy', 'peakef']}
            if Ngo == False:
                Ngo = {'run' : False}
            
            # 2.a Get subjects
            subs = self.subs
            if isinstance(subs, list):
                None
            elif subs == 'all':
                    subs = next(walk(self.xml_dir))[1]
            else:
                logger.info('')
                logger.critical("'subs' must either be an array of Participant IDs or = 'all' ")
                return
            subs.sort()

            # 3.a. Loop over event types
            logger.info('') 
            for e, event in enumerate(evt_name):  
                # 3.b. Loop over subjects
                for i, sub in enumerate(subs):
                    
                    # 2.b. Get sessions
                    sessions = self.sessions
                    if sessions == 'all':
                        try:
                            sessions = next(walk(f'{self.xml_dir}/{sub}'))[1]  # Get list of session directories
                        except StopIteration:
                            logger.warning(f'No visits found in {self.xml_dir}{sub}. Skipping..')
                            continue
                    elif not isinstance(sessions, list):
                        logger.info('')
                        logger.critical("'sessions' must either be a list of Session IDs or = 'all'")
                        return
                    
                    if not sessions:
                        logger.warning(f'No visits found in {self.xml_dir}{sub}. Skipping..')
                        continue
                    
                    # 3.c. Loop over sessions
                    for v, ses in enumerate(sessions):  
                        logger.debug(f'Extracting {event} parameters for {sub}, {ses}..') 
                        
                        # 3.d. Get recording
                        rdir = f'{self.rec_dir}/{sub}/{ses}/eeg/'
                        edf_file = [x for x in listdir(rdir) if x.endswith(filetype)
                                    if not x.startswith('.')]
                        
                        # 3.d. Get proper annotations file
                        if keyword is not None:
                            xml_file = [x for x in listdir(self.xml_dir + '/' + sub + '/' + ses) 
                                    if x.endswith('.xml') if not x.startswith('.') if keyword in x] 
                        else:
                            xml_file = [x for x in listdir(self.xml_dir +  '/' + sub + '/' + ses ) 
                                    if x.endswith('.xml') if not x.startswith('.')]    
                        if len(xml_file) == 0:                
                            logger.warning(f'{event} has not been detected for {sub}, {ses} - skipping..')
                            flag+=1
                            continue
                        elif len(xml_file) > 1:
                            logger.warning(f"More than 1 annotations file found for {sub}, visit {ses} - to select the correct file you must define the variable 'keyword' - skipping..")
                            flag+=1
                            continue
                        else:
                            xml_file_path = f'{self.xml_dir}/{sub}/{ses}/{xml_file[0]}'
                            
                            # 4.a. Open dataset
                            dataset = Dataset(rdir + edf_file[0])
                    
                            # 4.b. Import Annotations file
                            annot = Annotations(xml_file_path, rater_name=self.rater)
                        
                            # 4.c. Get sleep cycles (if any)
                            if cycle_idx is not None:
                                all_cycles = annot.get_cycles()
                                if all_cycles:
                                    cycle = [all_cycles[y - 1] for y in cycle_idx if y <= len(all_cycles)]
                                else:
                                    logger.warning(f'Cycles have not been marked for {sub}, {ses} - exporting all cycles..')
                                    flag+=1
                                    cycle = None
                            else:
                                cycle = None
                            
                            # 4.d. Channel setup
                            flag, chanset = load_channels(sub, ses, self.chan, 
                                                          self.ref_chan, flag, logger)
                            if not chanset:
                                continue
                            
                            newchans = rename_channels(sub, ses, self.chan, logger)
                            
                            # 5. Run through channels
                            for ch, channel in enumerate(chanset):
                                
                                # 5.a. Define full channel name (in annotations file)
                                if Ngo['run'] == False:
                                    if self.grp_name is not None:
                                        chan_ful = [f'{channel} ({self.grp_name})']
                                    else:
                                        chan_ful = [channel]
                                        logger.debug(f'channel is {chan_ful}')
                                else:
                                    chan_ful = Ngo['chan']
                                if type(chan_ful)!=str and len(chan_ful)>1:
                                    chan_ful = chan_ful[0]
                                    
                                # 5.b Rename channel for output file (if required)
                                if newchans:
                                    fnamechan = newchans[channel]
                                else:
                                    fnamechan = channel
                                
                                # 5.c. Define frequency bands
                                if adap_bands == 'Fixed':
                                    logger.debug("Adapted bands has been set as 'Fixed'.")
                                    freq = frequency
                                    if frequency == None:
                                        logger.critical("If setting adap_bands = 'Fixed' a frequency range must also be set.")
                                        logger.info('                      Check documentation for how to extract event parameters:')
                                        logger.info('                      https://seapipe.readthedocs.io/en/latest/index.html')
                                        return
                                elif adap_bands == 'Manual':
                                    logger.debug("Adapted bands has been set as 'Manual'. Will search for peaks within the tracking sheet")
                                    freq = read_manual_peaks(self.rootpath, sub, ses, channel, 
                                                             adap_bw, logger)
                                    if not freq:
                                        flag += 1
                                elif adap_bands == 'Auto':
                                    stagename = '-'.join(self.stage)
                                    if frequency == None:
                                        logger.critical("If setting adap_bands = 'Auto' a frequency range must also be set.")
                                        logger.info('                      Check documentation for how to extract event parameters:')
                                        logger.info('                      https://seapipe.readthedocs.io/en/latest/index.html')
                                        return
                                    else:
                                        band_limits = f'{frequency[0]}-{frequency[1]}Hz'
                                        logger.debug(f"Adapted bands has been set as 'Auto'. Will search for peaks within the limits: {band_limits}")
                                        freq = load_adap_bands(self.tracking['fooof'], sub, ses,
                                                               fnamechan, stagename, band_limits, 
                                                               adap_bw, logger)
                                
                                # Final freq check (and set defaults if none)
                                if freq == None and self.out_dir.split('/')[-1] == 'spindle':
                                    freq = (11,16)
                                elif freq == None and self.out_dir.split('/')[-1] == 'slowwave':
                                    freq = (0.5, 1.25)
                                elif freq == None:
                                    freq = (0, 35)
                                
                                band_limits = f'{round(freq[0],2)}-{round(freq[1],2)}Hz'
                                logger.debug(f"Using band limits: {band_limits}")
                                
                                # 5.d. Select and read data
                                logger.info('')
                                logger.debug(f"Reading data for {sub}, {ses}, channel {str(channel)}:{'-'.join(chanset[channel])}")
                                
                                # 5.e. Events from annotations file
                                evts = annot.get_events(name=event, time=None,
                                                        chan=f'{channel} ({self.grp_name})', 
                                                        stage=self.stage)
                                
                                if len(evts) == 0:
                                    logger.info('')
                                    logger.warning(f"Events: '{event}' haven't been detected for {sub}, {ses} on channel {channel}, skipping...")
                                
                                ### WHOLE NIGHT ###
                                elif model == 'whole_night':
                                    try:
                                        segments = fetch(dataset, annot, cat=cat, 
                                                         evt_type=[event], 
                                                         cycle=cycle, chan_full=chan_ful, 
                                                         reject_epoch=True, 
                                                         reject_artf = ['Artefact', 'Arou', 'Arousal'],)
                                        segments.read_data([channel], chanset[channel], 
                                                           grp_name=self.grp_name)
                                    except Exception as e:
                                        logger.error(e)
                                        logger.warning(f'Error reading data for {sub}, {ses}, CHANNEL {channel}.')
                                        flag +=1
                                        continue
                                    if len(segments) < 1:
                                        logger.warning(f'No valid data found for {sub}, {ses}, CHANNEL {channel}.')
                                        flag +=1
                                        continue
                                    else:
                                        # Get times
                                        poi = get_times(annot, stage=self.stage, 
                                                        cycle=cycle, 
                                                        chan=[channel], 
                                                        exclude=True)
                                        total_dur = sum([x[1] - x[0] for y in poi for x in y['times']])
                                        count = len(evts)
                                        density = count / (total_dur / epoch_dur)
                                        logger.debug('----- WHOLE NIGHT -----')
                                        logger.debug(f'No. Events = {count}, Total duration (s) = {total_dur}')
                                        logger.debug(f'Density = {round(density, ndigits=2)} per epoch')
                                        logger.info('')
                                        # Export event parameters
                                        try:
                                            data = event_params(segments, params=param_keys, 
                                                                band=freq, n_fft=None)
                                            if not path.exists(self.out_dir + '/' + sub):
                                                mkdir(self.out_dir + '/' + sub)
                                            if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                                                mkdir(self.out_dir + '/' + sub + '/' + ses)
                                            out_dir = self.out_dir + '/' + sub + '/' + ses
                                            data = sorted(data, key=lambda x: x['start'])
                                            stagename = '-'.join(self.stage)
                                            outputfile = f'{out_dir}/{sub}_{ses}_{fnamechan}_{stagename}_{event}.csv'
                                            export_event_params(outputfile, data, count=len(evts), 
                                                                density=density)
                                            logger.debug('Writing to ' + outputfile)
                                        except:
                                            logger.warning(f'Issue exporting data for {sub}, {ses}, {fnamechan}, {stagename}, {event}.')
                                            flag +=1
                                            continue
                                
                                ### PER STAGE AND CYCLE ###
                                elif model == 'stage*cycle':
                                    for s, st in enumerate(self.stage):
                                        for cy, cycc in enumerate(cycle_idx):
                                            try:
                                                cyc = cycle[cy]
                                                if not isinstance(chan_ful, list):
                                                    chan_ful = [chan_ful]
                                                    if len(chan_ful) > 1:
                                                        chan_ful = chan_ful[0] 
                                                segments = fetch(dataset, annot, 
                                                                 cat=cat, evt_type=[event], 
                                                                 stage = [st], 
                                                                 cycle=[cyc], 
                                                                 chan_full=chan_ful, 
                                                                 reject_epoch=True, 
                                                                 reject_artf=['Artefact', 'Arou', 'Arousal'], 
                                                                 min_dur=0.5)
                                                segments.read_data([channel], chanset[channel], 
                                                                   grp_name=self.grp_name)
                                            except Exception as e:
                                                logger.error(e)
                                                logger.warning(f'Error reading data for {sub}, {ses}, CHANNEL {channel}, STAGE {st} in CYCLE {cy+1}.')
                                                flag +=1
                                                continue
                                            if len(segments) < 1:
                                                logger.warning(f'No valid data found for {sub}, {ses}, CHANNEL {channel}, STAGE {st} in CYCLE {cy+1}.')
                                                flag +=1
                                                continue
                                            else:
                                                # Get times
                                                if isinstance(chan_ful, ndarray):
                                                    if len(chan_ful) > 1:
                                                        chan_ful = chan_ful[0]
            
                                                if len(chan_ful) > 1:
                                                    chan_ful = chan_ful[0]
            
                                                # Calculate event density (per cycle)
                                                poi = get_times(annot, stage=[st], 
                                                                cycle=[cyc], chan=[channel], 
                                                                exclude=True)
                                                total_dur = sum([x[1] - x[0] for y in poi for x in y['times']])
                                                evts = annot.get_events(name=event, 
                                                                        time=cycle[cy][0:2], 
                                                                        chan = f'{channel} ({self.grp_name})', 
                                                                        stage = st)
                                                count = len(evts)
                                                density = len(evts) / (total_dur / epoch_dur)
                                                logger.info('')
                                                logger.debug(f'---- STAGE {st}, CYCLE {cy+1} ----')
                                                logger.debug(f'No. Events = {count}, Total duration (s) = {total_dur}')
                                                logger.debug(f'Density = {round(density, ndigits=2)} per epoch')
                                                logger.info('')
    
                                                # Export event parameters
                                                data = event_params(segments, params=params, 
                                                                    band=freq, n_fft=None)
                                                if not path.exists(self.out_dir + '/' + sub):
                                                    mkdir(self.out_dir + '/' + sub)
                                                if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                                                    mkdir(self.out_dir + '/' + sub + '/' + ses)
                                                out_dir = self.out_dir + '/' + sub + '/' + ses
                                                data = sorted(data, key=lambda x: x['start'])
                                                outputfile = f'{out_dir}/{sub}_{ses}_{fnamechan}_{st}_cycle{cy+1}_{event}.csv' 
                                                export_event_params(outputfile, data, 
                                                                    count=len(evts), 
                                                                    density=density)
                                                logger.debug('Writing to ' + outputfile)                
                                ### PER STAGE ###
                                elif model == 'per_stage':
                                    for s, st in enumerate(self.stage):
                                        try:
                                            segments = fetch(dataset, annot, cat=cat, 
                                                         evt_type=[event], 
                                                         stage = [st], cycle=cycle, 
                                                         chan_full=chan_ful, 
                                                         reject_epoch=True, 
                                                         reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                                                         min_dur=0.5)
                                            segments.read_data([channel], chanset[channel], 
                                                               grp_name=self.grp_name)
                                        except Exception as e:
                                            logger.error(e)
                                            logger.warning(f'Error reading data for {sub}, {ses}, CHANNEL {channel}, STAGE {st}.')
                                            flag +=1
                                            continue
                                        if len(segments) < 1:
                                            logger.warning(f'No valid data found for {sub}, {ses}, CHANNEL {channel}, STAGE {st}.')
                                            flag +=1
                                            continue
                                        else:
                                            # Get times
                                            poi = get_times(annot, stage=[st], cycle=cycle, 
                                                            chan=[channel], exclude=True)
                                            total_dur = sum([x[1] - x[0] for y in poi for x in y['times']])
                                            evts = annot.get_events(name=event, time=None, 
                                                                    chan = f'{channel} ({self.grp_name})', 
                                                                    stage = st)
                                            count = len(evts)
                                            density = len(evts) / (total_dur / epoch_dur)
                                            logger.info('')
                                            logger.debug(f'---- STAGE {st} ----')
                                            logger.debug(f'No. Events = {count}, Total duration (s) = {total_dur}')
                                            logger.debug(f'Density = {round(density, ndigits=2)} per epoch')
                                            logger.info('')
                                            # Export event parameters 
                                            data = event_params(segments, params=params, 
                                                                band=freq, n_fft=None)
                                            if not data:
                                                data = [x for x in segments]
                                            else:
                                                data = sorted(data, key=lambda x: x['start'])
                                            
                                            if not path.exists(self.out_dir + '/' + sub):
                                                mkdir(self.out_dir + '/' + sub)
                                            if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                                                mkdir(self.out_dir + '/' + sub + '/' + ses)
                                            out_dir = self.out_dir + '/' + sub + '/' + ses
                                            
                                            outputfile = f'{out_dir}/{sub}_{ses}_{fnamechan}_{st}_{event}.csv' 
                                            export_event_params(outputfile, data, 
                                                                count=len(evts), 
                                                                density=density)
                                            logger.debug('Writing to ' + outputfile) 
                                
                                ### PER CYCLE ###
                                elif model == 'per_cycle': 
                                    for cy, cycc in enumerate(cycle_idx):
                                        try:
                                            cyc = cycle[cy]
                                            if not isinstance(chan_ful, list):
                                                chan_ful = [chan_ful]
                                                if len(chan_ful) > 1:
                                                    chan_ful = chan_ful[0] 
                                            segments = fetch(dataset, annot, cat=cat, 
                                                             evt_type=[event],
                                                             stage = self.stage, 
                                                             cycle=[cyc], chan_full=chan_ful, 
                                                             reject_epoch=True, 
                                                             reject_artf = ['Artefact', 'Arou', 'Arousal'],
                                                             min_dur=0.5)
                                            segments.read_data([channel], chanset[channel], 
                                                               grp_name=self.grp_name)
                                        except Exception:
                                            logger.error(e)
                                            logger.warning(f'Error reading data for {sub}, {ses}, CHANNEL {channel}, CYCLE {cy}.')
                                            flag +=1
                                            continue
                                        if len(segments) < 1:
                                            logger.warning(f'No valid data found for {sub}, {ses}, CHANNEL {channel}, CYCLE {cy}.')
                                            flag +=1
                                            continue
                                        else:
                                            # Get times
                                            if isinstance(chan_ful, ndarray):
                                                if len(chan_ful) > 1:
                                                    chan_ful = chan_ful[0]
        
                                            if len(chan_ful) > 1:
                                                chan_ful = chan_ful[0]
        
                                            # Calculate event density (per cycle)
                                            poi = get_times(annot, stage=self.stage, 
                                                            cycle=[cyc], chan=[channel], 
                                                            exclude=True)
                                            total_dur = sum([x[1] - x[0] for y in poi for x in y['times']])
                                            evts = annot.get_events(name=event, 
                                                                    time=cycle[cy][0:2], 
                                                                    chan = f'{channel} ({self.grp_name})', 
                                                                    stage = self.stage)
                                            count = len(evts)
                                            density = len(evts) / (total_dur / epoch_dur)
                                            logger.info('')
                                            logger.debug(f'---- CYCLE {cy+1} ----')
                                            logger.debug(f'No. Events = {count}, Total duration (s) = {total_dur}')
                                            logger.debug(f'Density = {round(density, ndigits=2)} per epoch')
                                            logger.info('')
                                            # Export event parameters 
                                            data = event_params(segments, params=params, 
                                                                band=freq, n_fft=None)
                                            if not path.exists(self.out_dir + '/' + sub):
                                                mkdir(self.out_dir + '/' + sub)
                                            if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                                                mkdir(self.out_dir + '/' + sub + '/' + ses)
                                            out_dir = self.out_dir + '/' + sub + '/' + ses
                                            data = sorted(data, key=lambda x: x['start'])
                                            stagename = '-'.join(self.stage)
                                            outputfile = f'{out_dir}/{sub}_{ses}_{fnamechan}_{stagename}_cycle{cy+1}_{event}.csv'  
                                            export_event_params(outputfile, data, 
                                                                count=len(evts), 
                                                                density=density)
                                            logger.debug('Writing to ' + outputfile)
                                    
                                ### PER SEGMENT ###
                                if segs:
                                    segnames = ['-'.join((x[0],x[1])) for x in segs]
                                    for s,seg in enumerate(segs): 
                                        
                                        # Calculate event density (per segment type)
                                        poi = get_times(annot, evt_type=[seg[0],seg[1]], 
                                                        stage=None, chan=[channel], 
                                                        exclude=True)
                                        duos=[]
                                        [duos.extend([(poi[0]['times'][x][0],poi[1]['times'][x][1])]) 
                                                             for x,item in enumerate(poi[0]['times'])]
                                        total_dur = sum([x[1] - x[0] for x in duos])
                                        evts =[]
                                        for d in duos:
                                            evts.extend(annot.get_events(name=event, 
                                                                         time=d, 
                                                                         chan = chan_ful, 
                                                                         stage = None))
                                        count = len(evts)
                                        density = len(evts) / (total_dur / epoch_dur)
                                        logger.info('')
                                        logger.debug(f'----- Segment {seg} -----')
                                        logger.debug(f'No. Events = {count}, Total duration (s) = {total_dur}')
                                        logger.debug(f'Density = {round(density, ndigits=2)} per epoch')
                                        logger.info('')
                                        
                                        # Export event parameters
                                        data = event_params(segments, params=params, 
                                                            band=freq, n_fft=None)
                                        if data:
                                                data = sorted(data, key=lambda x: x['start'])
                                                outputfile = f'{self.out_dir}/{sub}/{ses}/{sub}_{ses}_{fnamechan}_{segnames[s]}_{event}.csv'
                                                export_event_params(outputfile, data, 
                                                                    count=len(evts), 
                                                                    density=density)
                                        else:
                                            logger.warning(f'No valid data found for {sub}, {ses}, {seg}.')
                                            flag+=1
                                            continue

                ### 3. Check completion status and print
                if flag == 0:
                    logger.info('')
                    logger.debug('Listing Individual aNnotated Events (L.I.N.E) '
                                 'finished without ERROR.')  
                else:
                    logger.info('')
                    logger.warning('Listing Individual aNnotated Events (L.I.N.E) ' 
                                   f'finished with {flag} WARNINGS. See log for details.')
                return 
    
                    
    def net(self, chan, evt_name, adap_bands, params = 'all', cat = (1,1,1,1), 
                  cycle_idx = None, logger = create_logger('Event dataset')):
        
        '''
            aNnotated Event Tabulation (NET)
            
            This function extracts average (or stdev) parameters of 
            specific (named) events from the whole cohort and creates a 
            master-level dataframe tabulating this information.
            This function can only be used one-event-at-a-time.
        '''
        
        ### 0.a Set up logging
        flag = 0
        logger.info('')
        logger.debug(r""" Commencing Event Dataset Creation

                     
                              ∆       ∆       ∆    ∆     ∆       ∆    ∆         ∆
                      ~^~^~~^~O~^~^~^~O~^~^~^~O~^~~O~^~^~O~^~^~^~O~^~~O~^~^~^~^~O~^~^~~
                              .       .       .    .     .       .    .         .
                               •. . .• •..• .• •..• •. .• •..• .• •..• •...•.•.•.
                               •. •.•   ;   ;   ; »<>;  ;   ;   ;   ;   ;   ;  .                                                                                                  
                                 •.•....;...;...;...;...;...;...;...;...;...;.•
                      »<>          .•   ;   ;   ;   ;   ;   ;   ;  ;   ;    ;  »<>
                                   •....;...;...;...;...;...;...;...;...;...;.
                                   .•   ;   ;   ;   ;  ; »<> ;   ;  ;   ;    ;..
                                  .•....;...;...;...;...;...;...;...;...;...;..•.
                                   •.   ;  ;    ;   ;   ;   ;   ;  ;   ;    ;   ;
                                   .•..•..;.•...•...•...•..•..;•..•...•....•.;.•
                                
                                
                                
                                aNnotated Events Tabulation 
                                (N.E.T) 
                                
                                                    """,)
        
        ### 1. First check the directories
        # a. Check for output folder, if doesn't exist, create
        if not path.exists(self.out_dir):
                mkdir(self.out_dir)
        
        # b. Get subject IDs
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = next(walk(self.xml_dir))[1]
        else:
            logger.error("'subs' must either be an array of participant ids or = 'all' ")
            
        # c. Take a look through directories to get sessions
        subs.sort()
        sessions = {}
        for s, sub in enumerate(subs):
            if self.sessions == 'all':
                sessions[sub] = next(walk(f'{self.xml_dir}/{sub}'))[1]
        sessions = list(set([y for x in sessions.values() for y in x]))           
        sessions.sort() 
        
        ### 2. Set up organisation of export
        if cat[0] + cat[1] == 2:
            model = 'whole_night'
            logger.debug('Exporting parameters for the whole night.')
        elif cat[0] + cat[1] == 0:
            model = 'stage*cycle'
            logger.debug('Exporting parameters per stage and cycle separately.')
        elif cat[0] == 0:
            model = 'per_cycle'
            logger.debug('Exporting parameters per cycle separately.')
        elif cat[1] == 0:
            model = 'per_stage'  
            logger.debug('Exporting parameters per stage separately.')
        if 'cycle' in model and cycle_idx == None:
            logger.info('')
            logger.critical("To export cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
            return
        
        # 3. Set variable names and combine with visits 
        if params == 'all':
            variables = ['Count','Density','Duration_mean','Duration_stdv',
                         'Min_amplitude_mean','Min_amplitude_stdv', 'Max_amplitude_mean',
                         'Max_amplitude_stdv','Ptp_amplitude_mean', 'Ptp_amplitude_stdev',
                         'Power_mean','Power_stdev', 'Peak_power_frequency_mean',
                         'Peak_power_frequency_std']
        else:
            variables = params
        
        
        # 4. Begin data extraction
        for c, ch in enumerate(chan):
            logger.debug(f'Creating a {evt_name} dataset for {ch}..')
            
            # a. Create column names (append chan and ses names)
            for v, ses in enumerate(sessions):
                sesvar = []
                for pair in product(variables, [ses]):
                    sesvar.append('_'.join(pair))
                columns = []
                for pair in product([evt_name], sesvar, [ch]):
                    columns.append('_'.join(pair))
                
                # b. Extract data based on cycle and stage setup
                if model == 'whole_night':
                    stagename = '-'.join(self.stage)
                    logger.debug(f'Collating {evt_name} parameters from {ch}, {stagename}..')
                    st_columns = [x + f'_{stagename}' for x in columns]
                    df = DataFrame(index=subs, columns=st_columns,dtype=float) 
                    for s, sub in enumerate(subs): 
                        logger.debug(f'Extracting from {sub}, {ses}')
                        data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{evt_name}.csv' 
                        if path.isfile(data_file):
                            try:
                                df.loc[sub] = extract_event_data(data_file, variables)
                            except:
                                extract_data_error(logger)
                                flag +=1
                                return
                        else:
                            flag +=1
                            logger.warning(f'Data not found for {sub}, {ses}, '
                                           f'{ch}, {stagename}, Event: {evt_name} '
                                           '- has export_eventparams been run for '
                                           f'{model}, using adap_bands = {adap_bands}?')
                    if not path.exists(f'{self.out_dir}/{evt_name}_{model}'):
                        mkdir(f'{self.out_dir}/{evt_name}_{model}')
                    df.to_csv(f"{self.out_dir}/{evt_name}_{model}/{evt_name}_{ses}_{ch}_{stagename}.csv")
                    
                elif model == 'stage*cycle':
                    for cyc in cycle_idx:
                        cycle = f'cycle{cyc}'
                        for st in self.stage:
                            st_columns = [x + f'_{st}_{cycle}' for x in columns]
                            df = DataFrame(index=subs, columns=st_columns,dtype=float) 
                            logger.debug(f'Collating {evt_name} parameters from {ch}, {st}..')
                            for s, sub in enumerate(subs): 
                                logger.debug(f'Extracting from {sub}, {ses}')
                                data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{st}_{cycle}_{evt_name}.csv'
                                if path.isfile(data_file):
                                    try:
                                        df.loc[sub] = extract_event_data(data_file, variables)
                                    except:
                                        extract_data_error(logger)
                                        flag +=1
                                        return
                                else:
                                    flag +=1
                                    logger.warning(f'Data not found for {sub}, '
                                                   f'{ses}, {ch}, {st}, Event: '
                                                   f'{evt_name} - has export_eventparams '
                                                   f'been run for {model}, '
                                                   f'using adap_bands = {adap_bands}?')
                            if not path.exists(f'{self.out_dir}/{evt_name}_{model}'):
                                mkdir(f'{self.out_dir}/{evt_name}_{model}')
                            df.to_csv(f"{self.out_dir}/{evt_name}_{model}/{evt_name}_{ses}_{ch}_{st}_{cycle}.csv")
                
                elif model == 'per_cycle':
                    for cyc in cycle_idx:
                        cycle = f'cycle{cyc}'
                        stagename = '-'.join(self.stage)
                        st_columns = [x + f'_{stagename}_{cycle}' for x in columns]
                        df = DataFrame(index=subs, columns=st_columns,dtype=float) 
                        logger.debug(f'Collating {evt_name} parameters from {ch}, '
                                     f'{stagename}, {cycle}..')
                        for s, sub in enumerate(subs): 
                            logger.debug(f'Extracting from {sub}, {ses}')
                            data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{cycle}_{evt_name}.csv'
                            if path.isfile(data_file):
                                try:
                                    df.loc[sub] = extract_event_data(data_file, variables)
                                except:
                                    extract_data_error(logger)
                                    flag +=1
                                    return
                            else:
                                flag +=1
                                logger.warning(f'Data not found for {sub}, {ses}, '
                                               f'{ch}, {stagename}, {cycle}, Event: '
                                               f'{evt_name} - has export_eventparams '
                                               f'been run for {model}, using adap_bands '
                                               f'= {adap_bands}?')
                        if not path.exists(f'{self.out_dir}/{evt_name}_{model}'):
                            mkdir(f'{self.out_dir}/{evt_name}_{model}')
                        df.to_csv(f"{self.out_dir}/{evt_name}_{model}/{evt_name}_{ses}_{ch}_{stagename}_{cycle}.csv")
                    
                elif model == 'per_stage':
                    for st in self.stage:
                        st_columns = [x + f'_{st}' for x in columns]
                        df = DataFrame(index=subs, columns=st_columns,dtype=float) 
                        logger.debug(f'Collating {evt_name} parameters from {ch}, {st}..')
                        for s, sub in enumerate(subs): 
                            logger.debug(f'Extracting from {sub}, {ses}')
                            data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{st}_{evt_name}.csv'
                            if path.isfile(data_file):
                                try:
                                    df.loc[sub] = extract_event_data(data_file, variables)
                                except:
                                    extract_data_error(logger)
                                    flag +=1
                                    return
                            else:
                                flag +=1
                                logger.warning(f'Data not found for {sub}, {ses}, {ch}, {st}, '
                                               f'Event: {evt_name} - has export_eventparams been '
                                               f'run for {model}, using adap_bands = {adap_bands}?')
                        if not path.exists(f'{self.out_dir}/{evt_name}_{model}'):
                            mkdir(f'{self.out_dir}/{evt_name}_{model}')  
                        df.to_csv(f"{self.out_dir}/{evt_name}_{model}/{evt_name}_{ses}_{ch}_{st}.csv")
        
                
        ### 3. Check completion status and print
        if flag == 0:
            logger.info('')
            logger.debug('aNnotated Events Tabulation (N.E.T) finished without ERROR.')  
        else:
            logger.info('')
            logger.warning('aNnotated Events Tabulation (N.E.T) finished with '
                           f'{flag} WARNINGS. See log for details.')
        return 

    
    def scalops(self, chan, evt_name, bands, params = 'all', 
                 logger = create_logger('Cluster dataset')):
        
        '''
            Summary of Clustering And Low-frequency Oscillatory Power Signatures
            S.C.A.L.O.P.S
            
            This function extracts individual parameters of 
            clustering and LFF stats from the whole cohort and creates a 
            master-level dataframe tabulating this information.
            This function can only be used one-event-at-a-time.
        '''
        
        ### 0.a Set up logging
        flag = 0
        logger.info('')
        logger.debug(r""" Commencing Event Dataset Creation

                     
                               _.---._              _.---._
                           .'"".'/|\`.""'.      .'""/     \""'. 
                          :  .' / | \ `.  :    :  .'\-___-/`.  :
                          '.'  /  |  \  `.'    '.'  /  |  \  `.'
                           `. /   |   \ .'      `. /   |   \ .'
                             `-.__|__.-'          `-.__|__.-'
                                                    
                    Summary of Clustering And Low-frequency Oscillatory Power Signatures
                    S.C.A.L.O.P.S
                                
                                                    """,)
        
        ### 1. First check the directories
        # a. Check for output folder, if doesn't exist, create
        if not path.exists(self.out_dir):
                mkdir(self.out_dir)
        
        # b. Get subject IDs
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = next(walk(self.xml_dir))[1]
        else:
            logger.error("'subs' must either be an array of participant ids or = 'all' ")
            
        # c. Take a look through directories to get sessions
        subs.sort()
        if isinstance(self.sessions, str) and self.sessions == "all":
            sessions = set()
            for subdir in Path(self.xml_dir).iterdir():
                if subdir.is_dir() and subdir.name.startswith("sub-"):
                    for sesdir in subdir.iterdir():
                        if sesdir.is_dir() and sesdir.name.startswith("ses-"):
                            sessions.add(sesdir.name)
            sessions = sorted(sessions)
        elif isinstance(self.sessions, list):
            sessions = self.sessions
        else:
            raise ValueError("sessions must be a list or 'all'")
 
        # 3. Set variable names and combine with visits 
        if params == 'all':
            variables = ['cv','cv_p_value','VMR','VMR_p_value',
                         'num_clusters_n','average_cluster_size_n',
                         'max_cluster_size_n','min_cluster_size_n',
                         'clusters_per_60s','ave_within_cluster_freq',
                         'mean_iei_s','median_iei_s','min_interval_s',
                         'mean_interval_s','max_interval_s',
                         'min_cluster_duration_s','avg_cluster_duration_s',
                         'max_cluster_duration_s']
            for band in bands:
                variables.extend([f'{band}_peak_freq', f'{band}_avg_peak_power'])
        else:
            variables = params
        
        # band names
        fbandnames = '-'.join(bands)
        
        # 4. Begin data extraction
        for c, ch in enumerate(chan):
            logger.debug(f'Collating cluster parameters for {fbandnames}, {evt_name} in {ch}..')
            # a. Create column names (append chan and ses names)
            for v, ses in enumerate(sessions):
                sesvar = []
                for pair in product(variables, [ses]):
                    sesvar.append('_'.join(pair))
                columns = []
                for pair in product([evt_name], sesvar, [ch]):
                    columns.append('_'.join(pair))
                
                stagename = '-'.join(self.stage)
                st_columns = [x + f'_{stagename}' for x in columns]
                df_stats = DataFrame(index=subs, columns=st_columns, dtype=float)
                df_hist = []
                for s, sub in enumerate(subs): 
                    logger.debug(f'Extracting from {sub}, {ses}')
                    
                    # Cluster metrics
                    ar = []
                    data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{evt_name}_clustering.csv'     
                    if path.isfile(data_file):
                        try:
                            ar.extend(extract_cluster_data(data_file, variables))
                        except:
                            extract_data_error(logger)
                            flag +=1
                            continue
                    else:
                        flag +=1
                        logger.warning(f'*clustering.csv not found for {sub}, '
                                       f'{ses}, {ch}, {stagename}, {fbandnames}'
                                       f'in {self.xml_dir}/{sub}/{ses}/')
                        continue
                    
                    # Peak data
                    data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{fbandnames}_fluctuation_stats.csv'     
                    if path.isfile(data_file):
                        try:
                            ar.extend(extract_cluster_data(data_file, variables))
                        except:
                            extract_data_error(logger)
                            flag +=1
                            continue
                    else:
                        flag +=1
                        logger.warning(f'*fluctuation_stats.csv not found for {sub}, '
                                       f'{ses}, {ch}, {stagename}, {fbandnames} '
                                       f'in {self.xml_dir}/{sub}/{ses}/')
                        continue
                        
                    # Save to df
                    df_stats.loc[sub] = ar
                 
                df_stats.to_csv(f"{self.out_dir}/{evt_name}_{ses}_{ch}_{stagename}_cluster_stats.csv")
                
        for c, ch in enumerate(chan):
            for band in bands:
                logger.debug(f'Collating phase-histograms for {band}, {evt_name} in {ch}..')
                
                # a. Create column names (append chan and ses names)
                for v, ses in enumerate(sessions):
                    # Histogram
                    df_hist = []
                    for s, sub in enumerate(subs): 
                        logger.debug(f'Extracting from {sub}, {ses}')   
                        data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{evt_name}_{band}_coupling.csv'
                        if path.isfile(data_file):
                            try:
                                hist = read_csv(data_file, index_col = 0)
                                hist.index = [sub]
                                df_hist.append(hist)
                            except:
                                extract_data_error(logger)
                                flag +=1
                                continue
                        else:
                            flag +=1
                            logger.warning(f'*coupling.csv not found for {sub}, '
                                           f'{ses}, {ch}, {stagename}, {evt_name}, {band} '
                                           f'in {self.xml_dir}/{sub}/{ses}/')
                            
                            
                   
                    # Concatenate all dataframes vertically (row-wise)
                    master_df = concat(df_hist)
                    master_df.to_csv(f"{self.out_dir}/{evt_name}-{band}_{ses}_{ch}_{stagename}_histogram.csv")
            
        for c, ch in enumerate(chan):
            for band in bands:
                logger.debug(f'Collating PSD for {band}, {evt_name} in {ch}..')
                
                # a. Create column names (append chan and ses names)
                for v, ses in enumerate(sessions):
                    # Histogram
                    df_psd = []
                    for s, sub in enumerate(subs): 
                        logger.debug(f'Extracting from {sub}, {ses}')   
                        data_file = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{band}_fluctuations_psd.csv'
                        if path.isfile(data_file):
                            try:
                                psd = read_csv(data_file, index_col = 0)
                                psd.index = [sub]
                                df_psd.append(psd)
                            except:
                                extract_data_error(logger)
                                flag +=1
                                continue
                        else:
                            flag +=1
                            logger.warning(f'*fluctuations_psd.csv not found for {sub}, '
                                           f'{ses}, {ch}, {stagename}, {evt_name}, {band} '
                                           f'in {self.xml_dir}/{sub}/{ses}/')
                            
                            
                   
                    # Concatenate all dataframes vertically (row-wise)
                    master_df = concat(df_psd)
                    master_df.to_csv(f"{self.out_dir}/{band}_{ses}_{ch}_{stagename}_fluctuations_psd.csv")
                        
                  
        ### 3. Check completion status and print
        if flag == 0:
            logger.info('')
            logger.debug('Summary of Clustering And Low-frequency Oscillatory Power Signatures '
                         '(S.C.A.L.O.P.S) finished without ERROR.')  
        else:
            logger.info('')
            logger.warning('Summary of Clustering And Low-frequency Oscillatory Power Signatures '
                           f'(S.C.A.L.O.P.S) finished with {flag} WARNINGS. See log for details.')
        return            
                    
    
    def pac_summary(self, chan, evt_name = None, 
                          adap_bands_phase = 'Fixed', 
                          frequency_phase = (0.5, 1.25), 
                          adap_bands_amplitude = 'Fixed', 
                          frequency_amplitude = (11, 16),
                          adap_bw = 4,
                          params = 'all', 
                          cat = (1,1,1,1), 
                          cycle_idx = None, 
                          logger = create_logger('PAC dataset')):
        
        
        '''
            PAC summary
            
            This function extracts the summary parameters of PAC for each sub 
            and ses, and creates a master-level dataframe tabulating this 
            information. This function can only be used for one set of analyses
            at a time.
        '''
        
        ### ---- 0.a Set up logging ----
        missing_data_flag = 0
        
        logger.info('')
        logger.debug(r""" Commencing PAC Dataset Creation

                     
                      (Graphic Under Construction...)
                      
                      ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀  ⢀⣀⣀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⢀⣠⣴⠾⢻⣿⡟⠻⠶⢦⣤⣀⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
                        ⠀⠀⠀⠀⣀⣤⠾⠛⠉⠀⠀⣸⠛⣷⠀⠀⠀⠀⠉⠙⠻⠶⣦⣤⣀⠀⠀⠀⠀⠀
                        ⠀⠀⠐⠛⠋⠀⠀⠀⠀⠀⠀⠛⠀⠛⠂⠀⠀⠀⠀⠀⠀⠀⠀⠈⠙⠛⠒⠂⠀⠀
                        ⠀⢸⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡇⠀
                        ⠀⠀⠀⢠⣤⣤⣤⠀⠀⠀⠀⢠⣤⡄⢠⣤⡄⠀⠀⠀⠀⠀⠀⠀⠀⠀⡄⠀⠀⠀
                        ⠀⠀⠀⠈⠉⠉⠉⠀⠀⠀⠀⠸⠿⠇⠸⠿⠃⠀⠀⠀⠀⠀⠀⠀⠀⠀⡇⠀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢰⣶⡆⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⡇⠀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⡇⠀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠃⠀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⡀⠃⡀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠘⠋⠈⠛⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣀⢸⣿⡇⢀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣿⣿⣿⣿⡇⠀
                        ⠀⠀⠀⠀⠀⠀⠀⢀⣴⠾⠋⢸⣿⡇⠈⠳⣦⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
                        ⠀⠀⠀⠀⠀⠀⠀⠈⠁⠀⠀⠈⠛⠃⠀⠀⠀⠉⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
                                
                                                    """,)
        
        ### ---- 1. First check the directories ----

        # a. Ensure output directory exists
        if not path.exists(self.out_dir):
            mkdir(self.out_dir)
        
        # b. Get subject IDs
        subs = self.subs
        sub_pattern = re.compile(r"^sub-\w+")
        
        if isinstance(subs, list):
            if not all(sub_pattern.match(s) for s in subs):
                logger.error("Invalid participant ID format in list. IDs should "
                             "be BIDS-compatible (e.g., 'sub-01').")
                return
        elif subs == 'all':
            try:
                subs = sorted(next(walk(self.xml_dir))[1])
                subs = [s for s in subs if sub_pattern.match(s)]
            except StopIteration:
                logger.critical(f"No participant directories found in {self.xml_dir}")
                return
        else:
            logger.error("'subs' must either be a list of participant IDs or 'all'")
            return
        
        # c. Get all sessions (may differ across subjects)
        if isinstance(self.sessions, str) and self.sessions == "all":
            sessions = set()
            for subdir in Path(self.xml_dir).iterdir():
                if subdir.is_dir() and subdir.name.startswith("sub-"):
                    for sesdir in subdir.iterdir():
                        if sesdir.is_dir() and sesdir.name.startswith("ses-"):
                            sessions.add(sesdir.name)
            sessions = sorted(sessions)
        elif isinstance(self.sessions, list):
            sessions = self.sessions
        else:
            raise ValueError("sessions must be a list or 'all'")
        
        # ---- 2. Determine export model based on 'cat' ----    
        if not isinstance(cat, tuple) or len(cat) != 4:
            logger.critical("'cat' must be a tuple of four binary flags: (cycle, "
                            "stage, continuous, event).")
            return
        
        cycle_flag, stage_flag, _, _ = cat  # continuous and event flags are unused here
        
        if cycle_flag and stage_flag:
            model = 'whole_night'
            logger.debug('Exporting parameters for the whole night (concatenated '
                         'across cycles and stages).')
        elif not cycle_flag and not stage_flag:
            model = 'stage*cycle'
            logger.debug('Exporting parameters per stage and per cycle separately.')
        elif not cycle_flag:
            model = 'per_cycle'
            logger.debug('Exporting parameters per cycle separately (but across '
                         'stages).')
        elif not stage_flag:
            model = 'per_stage'
            logger.debug('Exporting parameters per stage separately (but across '
                         'cycles).')
        else:
            logger.critical("Unrecognized combination of 'cat' flags for cycle "
                            "and stage.")
            return
        
        # Validation: If we're splitting by cycle, we need to know which ones
        if 'cycle' in model and cycle_idx is None:
            logger.critical("To export data by cycle, 'cycle_idx' must be specified.")
            return
        
        
        # ---- 3. Set PAC variable names and column generation ----
        if params == 'all':
            variables = ['mi_raw', 'mi_norm', 'pval1', 'pp_radians', 
                         'ppdegrees', 'mvl']
            logger.debug(f"Using default PAC variables: {variables}")
        elif isinstance(params, list) and all(isinstance(p, str) for p in params):
            variables = params
            logger.debug(f"Using custom PAC variables: {variables}")
        else:
            logger.critical("'params' must be 'all' or a list of variable names (strings).")
            return


        # ---- 4. Set PAC band names for phase and amplitude ----
        # Phase Band Naming
        if adap_bands_phase == 'Fixed':
            adap_phase_name = 'fixed'
        else:
            adap_phase_name = 'adap'

        # Amplitude Band Naming
        if adap_bands_amplitude == 'Fixed':
            adap_amp_name = 'fixed'
        else:
            adap_amp_name = 'adap'
        
        # Generate pac_name with event if specified, otherwise without
        if evt_name:
            pac_name = f'{evt_name}_**-{adap_phase_name}_##-{adap_amp_name}_pac'
        else:
            pac_name = f'**-{adap_phase_name}_##-{adap_amp_name}_pac'
        
        # Setup for output filename
        out_phasebw = f'{frequency_phase[0]}-{frequency_phase[1]}Hz'  
        out_ampbw = f'{frequency_amplitude[0]}-{frequency_amplitude[1]}Hz' 
        pac_outname = f'pac_pha-{out_phasebw}-{adap_phase_name}_amp-{out_ampbw}-{adap_amp_name}'

        # ---- 5. Begin data extraction ----
        for c, ch in enumerate(chan):
            logger.debug(f'Creating a PAC dataset for {ch}..')
                
            for v, ses in enumerate(sessions):
                sesvar = []
                for pair in product(variables, [ses]):
                    sesvar.append('_'.join(pair))
                columns = []
                for pair in product(sesvar, [ch]):
                    columns.append('_'.join(pair))
                    
                missing_data_flag = extract_pac_summary(subs, 
                                                        ses, 
                                                        model,
                                                        evt_name,
                                                        pac_name,
                                                        pac_outname,
                                                        ch, 
                                                        self.stage, 
                                                        cycle_idx,
                                                        columns, 
                                                        variables, 
                                                        self.rootpath, 
                                                        self.xml_dir, 
                                                        self.out_dir,
                                                        adap_bands_phase,  
                                                        frequency_phase,
                                                        adap_bands_amplitude, 
                                                        frequency_amplitude, 
                                                        adap_bw, 
                                                        self.tracking, 
                                                        logger, 
                                                        missing_data_flag
                                                        )    
                    
        # ---- 6. Check completion status and print ----
        if missing_data_flag == 0:
            logger.info('')
            logger.debug('Create PAC dataset finished without ERROR.')  
        else:
            logger.info('')
            logger.warning(f'Create PAC dataset finished with {missing_data_flag} WARNINGS. '
                           'See log for details.')
        
        return 
       
    
def extract_pac_summary(subs, ses, model, evt_name, pac_name, pac_outname, 
                        chan, stage, cycle_idx, 
                        columns, variables, rootpath, xml_dir, out_dir, 
                        adap_bands_phase, frequency_phase,
                        adap_bands_amplitude, frequency_amplitude, 
                        adap_bw, 
                        tracking, logger, flag):
    
    
    if not path.exists(f'{out_dir}/{pac_outname}_{model}'):
        mkdir(f'{out_dir}/{pac_outname}_{model}')

    evt = f'_{evt_name}' if evt_name else ""
    
    # ---- Extract data based on cycle and stage setup ----
    if model == 'whole_night':
        stagename = '-'.join(stage)
        logger.debug(f'Collating PAC parameters from {chan}, {stagename}..')
        st_columns = [x + f'_{stagename}' for x in columns]
        df = DataFrame(index=subs, columns=st_columns, dtype=float) 
        out_filename = f'{ses}_{chan}_{stagename}{evt}_{pac_outname}.csv'
        for s, sub in enumerate(subs): 
            logger.debug(f'Extracting from {sub}, {ses}')
            
            phase_bw, amp_bw = extract_phase_amp_bw(rootpath, sub, ses, 
                                                           chan, stage, adap_bw,
                                                           adap_bands_phase, 
                                                           frequency_phase,
                                                           adap_bands_amplitude, 
                                                           frequency_amplitude,
                                                           tracking, logger)
            
            pac_name = pac_name.replace("**", phase_bw).replace("##", amp_bw)
            data_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_{chan}_{stagename}_{pac_name}_parameters.csv' 
            df.loc[sub] = extract_pac_data(data_file, model, variables, logger)
            flag += df.loc[sub].isna().any()
        df.to_csv(f"{out_dir}/{pac_outname}_{model}/{out_filename}")

    elif model == 'stage*cycle':
        for cyc in cycle_idx:
            cycle = f'cycle{cyc}'
            for st in stage:
                logger.debug(f'Collating PAC parameters from {chan}, {st}..')
                st_columns = [x + f'_{st}_{cycle}' for x in columns]
                df = DataFrame(index=subs, columns=st_columns,dtype=float) 
                out_filename = f'{ses}_{chan}_{st}_{cycle}{evt}_{pac_outname}.csv'
                for s, sub in enumerate(subs): 
                    logger.debug(f'Extracting from {sub}, {ses}')
                    
                    phase_bw, amp_bw = extract_phase_amp_bw(rootpath, sub, ses, 
                                                                   chan, stage, adap_bw,
                                                                   adap_bands_phase, 
                                                                   frequency_phase,
                                                                   adap_bands_amplitude, 
                                                                   frequency_amplitude,
                                                                   tracking, logger)
                    
                    pac_name = pac_name.replace("**", phase_bw).replace("##", amp_bw)
                    data_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_{chan}_{st}_{cycle}_{pac_name}_parameters.csv'
                    df.loc[sub] = extract_pac_data(data_file, model, variables, logger)
                    flag += df.loc[sub].isna().any()
                df.to_csv(f"{out_dir}/{pac_outname}_{model}/{out_filename}")
                    
    elif model == 'per_cycle':
        for cyc in cycle_idx:
            cycle = f'cycle{cyc}'
            stagename = '-'.join(stage)
            logger.debug(f'Collating PAC parameters from {chan}, {stagename}, {cycle}..')
            st_columns = [x + f'_{stagename}_{cycle}' for x in columns]
            df = DataFrame(index=subs, columns=st_columns, dtype=float) 
            out_filename = f'{ses}_{chan}_{stagename}_{cycle}{evt}_{pac_outname}.csv'
            for s, sub in enumerate(subs): 
                logger.debug(f'Extracting from {sub}, {ses}')
                
                phase_bw, amp_bw = extract_phase_amp_bw(rootpath, sub, ses, 
                                                               chan, stage, adap_bw,
                                                               adap_bands_phase, 
                                                               frequency_phase,
                                                               adap_bands_amplitude, 
                                                               frequency_amplitude,
                                                               tracking, logger)
                
                pac_name = pac_name.replace("**", phase_bw).replace("##", amp_bw)
                data_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_{chan}_{stagename}_{cycle}_{pac_name}_parameters.csv'
                df.loc[sub] = extract_pac_data(data_file, model, variables, logger)
                flag += df.loc[sub].isna().any()
            df.to_csv(f"{out_dir}/{pac_outname}_{model}/{out_filename}")

    elif model == 'per_stage':
        for st in stage:
            st_columns = [x + f'_{st}' for x in columns]
            logger.debug(f'Collating PAC parameters from {chan}, {st}..')
            df = DataFrame(index=subs, columns=st_columns, dtype=float) 
            out_filename = f'{ses}_{chan}_{st}_{pac_outname}.csv'
            for s, sub in enumerate(subs): 
                logger.debug(f'Extracting from {sub}, {ses}')
                
                phase_bw, amp_bw = extract_phase_amp_bw(rootpath, sub, ses, 
                                                               chan, stage, adap_bw,
                                                               adap_bands_phase, 
                                                               frequency_phase,
                                                               adap_bands_amplitude, 
                                                               frequency_amplitude,
                                                               tracking, logger)
                
                pac_name = pac_name.replace("**", phase_bw).replace("##", amp_bw)
                data_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_{chan}_{st}_{pac_name}_parameters.csv'
                df.loc[sub] = extract_pac_data(data_file, model, variables, logger)
                flag += df.loc[sub].isna().any()
            df.to_csv(f"{out_dir}/{pac_outname}_{model}/{out_filename}")

    return flag
 
    
def extract_phase_amp_bw(rootpath, sub, ses, ch, stage, adap_bw,
                            adap_bands_phase, frequency_phase, 
                            adap_bands_amplitude, frequency_amplitude,
                            tracking, logger = create_logger("Extract Phase Amplitude Bandwidths")):
    
    def get_band(band_type, method, freq_range, label):
        if method == 'Fixed':
            band = freq_range
        elif method == 'Manual':
            chansheet = read_tracking_sheet(rootpath, logger)
            oldchans = reverse_chan_lookup(sub, ses, chansheet, logger)
            fnamechan = oldchans[ch]
            band = read_manual_peaks(rootpath, sub, ses, fnamechan, adap_bw, logger)
        elif method == 'Auto':
            stagename = '-'.join(stage)
            band_limits = f'{freq_range[0]}-{freq_range[1]}Hz'
            band = load_adap_bands(tracking['fooof'], sub, ses, ch,
                                   stagename, band_limits, adap_bw, logger)
        else:
            raise ValueError(f"Unknown {label} method: {method}")
        return band

    # Get phase and amplitude frequency bands
    f_pha = get_band('phase', adap_bands_phase, frequency_phase, 'phase')
    if not f_pha:
        f_pha = frequency_phase
    f_amp = get_band('amplitude', adap_bands_amplitude, frequency_amplitude, 'amplitude')
    if not f_amp:
        f_amp = frequency_amplitude
    
    # Format names
    phase_bw = f'pha-{round(f_pha[0], 2)}-{round(f_pha[1], 2)}Hz'
    amp_bw = f'amp-{round(f_amp[0], 2)}-{round(f_amp[1], 2)}Hz'

    return phase_bw, amp_bw



def extract_data_error(logger = create_logger(" ")):
    logger.critical("Data extraction error: Check that all 'params' are written correctly.")
    logger.info('                      Check documentation for how event parameters need to be written:')
    logger.info('                      https://seapipe.readthedocs.io/en/latest/index.html')
    


def extract_pac_data(data_file, model, variables, logger = create_logger("Extract PAC data")):
    
    data = []
    if not path.exists(data_file):
        logger.warning(f"{data_file} not found. Has pac been run for {model}, "
                       "with these frequency bands?")
        data = full(len(variables), nan)
    else:
        try:
            # Read csv
            df = read_csv(data_file, header=0)
            
            # PAC variables
            if 'mi_raw' in variables:
                data.append(df['mi_raw'][0])
            if 'mi_norm' in variables:
                data.append(df['mi_norm'][0])
            if 'pval1' in variables:
                data.append(df['pval1'][0])              
            if 'pp_radians' in variables:
                data.append(df['pp_radians'][0])               
            if 'ppdegrees' in variables:
                data.append(df['ppdegrees'][0])
            if 'mvl' in variables:
                data.append(df['mvl'][0])					
            data = asarray(data)
        except:
            extract_data_error(logger)  
            data = full(len(variables), nan)

    return data
    
    
def extract_event_data(data_file, variables):
    
    # Delimiter
    data_file_delimiter = ','
    
    # The max column count a line in the file could have
    largest_column_count = 0
    
    # Loop the data lines
    with open(data_file, 'r') as temp_f:
        # Read the lines
        lines = temp_f.readlines()
    
        for l in lines:
            # Count the column count for the current line
            column_count = len(l.split(data_file_delimiter)) + 1
            
            # Set the new most column count
            largest_column_count = column_count if largest_column_count < column_count else largest_column_count
    
    # Generate column names (will be 0, 1, 2, ..., largest_column_count - 1)
    column_names = [i for i in range(0, largest_column_count)]
    
    # Read csv
    df = read_csv(data_file, header=None, delimiter=data_file_delimiter, 
                  names=column_names, index_col=0)
    
    data = []
    if 'Count' in variables:
        data.append(float(df.loc['Count'][1]))
    if 'Density' in variables:
       data.append(round(float(df.loc['Density'][1]),3))
       
       
    df = read_csv(data_file, skiprows=3, header=0, delimiter=data_file_delimiter, 
                   index_col=0)
    
    if 'Duration_mean' in variables:
        data.append(df['Duration (s)'].loc['Mean'])
    if 'Duration_stdv' in variables:
        data.append(df['Duration (s)'].loc['SD'])
    
    if 'Min_amplitude_mean' in variables:
        data.append(df['Min. amplitude (uV)'].loc['Mean'])
    if 'Min_amplitude_stdv' in variables:
        data.append(df['Min. amplitude (uV)'].loc['SD'])
    
    if 'Max_amplitude_mean' in variables:
        data.append(df['Max. amplitude (uV)'].loc['Mean'])
    if 'Max_amplitude_stdv' in variables:
        data.append(df['Max. amplitude (uV)'].loc['SD'])
        
    if 'Ptp_amplitude_mean' in variables:
        data.append(df['Peak-to-peak amplitude (uV)'].loc['Mean'])
    if 'Ptp_amplitude_stdev' in variables:
        data.append(df['Peak-to-peak amplitude (uV)'].loc['SD'])
        
    if 'Power_mean' in variables:
        data.append(df['Power (uV^2)'].loc['Mean'])
    if 'Power_stdev' in variables:
        data.append(df['Power (uV^2)'].loc['SD'])
        
    if 'Peak_power_frequency_mean' in variables:
        data.append( df['Peak power frequency (Hz)'].loc['Mean'])
    if 'Peak_power_frequency_std' in variables:
        data.append(df['Peak power frequency (Hz)'].loc['SD'])
    
    
    data = asarray(data)

    return data
    

def extract_cluster_data(data_file, variables):
    
    # Delimiter
    data_file_delimiter = ','
    
    # Read csv
    df = read_csv(data_file, index_col=0, delimiter=data_file_delimiter)
    
    data = []
    for col in df.columns:
        if col in variables:
            data.append(df.loc[0,col])
            
    return data

        
    
def trawls():
    
    '''
        Tabulating and Reordering Aggregated WhoLe Statistics (TRAWLS)
        
        Function to combine multiple datasets together into 1.
    '''
    
    return    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 25 12:07:36 2023

@author: nathancross
"""
from datetime import datetime, date
from os import listdir, mkdir, path, remove, walk
from pandas import DataFrame
from seapipe.events.fish import FISH
from seapipe.events.whales import whales
from seapipe.spectrum.psa import (Spectrum, default_epoch_opts, default_event_opts,
                     default_fooof_opts, default_filter_opts, default_frequency_opts, 
                     default_general_opts,default_norm_opts)
from seapipe.utils.squid import SQUID, gather_qc_reports
from seapipe.stats import sleepstats
from seapipe.utils.audit import (check_dataset, extract_channels, make_bids,
                        track_processing)
from seapipe.utils.logs import create_logger, create_logger_outfile
from seapipe.utils.load import (check_chans, read_tracking_sheet, 
                                select_input_dirs, select_output_dirs, load_stages)
from seapipe.utils.misc import adap_bands_setup

## TO DO:
#   - adapt load channels to be flexible for non-equivalent refsets and chansets
#   - related to the above, fix logging when using adapted bands but chan or ref_chan
#               are set to something other than None - e.g. psa.py, line 367
#   - add in log for detection whether auto, fixed or adapted bands was run
#   - add logging to save to output file (not implemented for all functions)
#   - update adapted bands in tracking.tsv
#   - fix discrepency between track WARNINGS and output in dataframe 
#   - update initial tracking to include spindles, slow_oscillation, cfc, power_spectrum
#   - update export sleepstats to export by stage/cycle separately
#   - possibility of cycle_idx = 'all'
#   - enable macro_dataset per sleep cycle
#   - enable downsampling of data

## FOR DOCUMENTATION:
#   - Clearly describe how chanset & refset works, ie. chanset = per chan, refset = ALL chans

class pipeline:
        
    """Contains specific information and allows the application of methods of 
    analysis, associated with a dataset. 

    Parameters
    ----------
    indir : str 
        name of the root level directory containing the BIDS organised data
        
    outfile : bool / str
        whether to save log of dataset audit to file. If False (default) - does
        not save any log file. If True - saves a log under the filepath 
        /derivatives/seapipe/audit/audit.csv. Else if a string to a filepath, 
        it will save the log under the filepath indicated in the string.
        
    filetype : str
        extension of file to search for. Default is '.edf' - but according to 
        BIDS convention other filetypes can be '.vhdr', '.vmrk', '.eeg' or '.set'

    Attributes
    ----------
    rootpath : str
        name of the root level directory
    datapath : str
        name of the directory containing the raw data (recordings and annotations)
    outpath : str
        name of the directory containing the output (analysed) data

    """
        
    def __init__(self, indir, tracking = False, outfile = False, 
                       filetype = '.edf'):
        
        self.rootpath = indir
        if path.exists(indir + '/DATA'):
            self.datapath = indir + '/DATA'
        else:
            self.datapath = indir + '/sourcedata'
        self.outpath = indir + '/derivatives'
        if not path.exists(self.outpath):
            mkdir(self.outpath)
        self.outfile = outfile
        if not path.exists(f'{self.outpath}/audit'):
            mkdir(f'{self.outpath}/audit')
        self.log_dir = self.outpath + '/audit/logs/'
        if not path.exists(self.log_dir):
            mkdir(self.log_dir)
        self.tracking = {}
        self.audit_init = check_dataset(self.rootpath, self.datapath, 
                                        self.outfile, filetype, tracking)
        
        if tracking:
            self.track(subs = 'all', ses = 'all', 
                   step = ['staging','spindle','slowwave','pac','sync','psa'],
                   show = False, log = False)
        
    
    #--------------------------------------------------------------------------
    '''
    MISCELLANEOUS FUNCTIONS
    
    audit -> Audits dataset structure for compatibility with seapipe analysis.
    
    list_dataset ->  Intended to walk from root directory through participant 
                        folders and list all participants and their files.
    
    track -> Tracks what seapipe processing or functions have already been applied
                to a dataset, with information on which channels and parameters 
                have been used.
                
    make_bids (beta) -> Transforms data from (some) data structures into the 
                            correct BIDS format compatible with use for seapipe.
                            
    extract_channels -> Extracts and lists which channels exist in the dataset. 

    load_stages -> Extracts stages from the BIDS formatted dataset, in which
                        staging has been listed in a file *acq-PSGScoring_events.tsv
    
    '''    
        
        
    def audit(self, outfile = False, tracking = False, filetype = '.edf'):
        
        ''' Audits the dataset for BIDS compatibility.
            Includes option to save the audit to an output file.
        '''
        
        # Create audit directory
        out_dir = f'{self.outpath}/audit'
        if not path.exists(out_dir):
            mkdir(out_dir)
            
        if not outfile and not self.outfile:
            logger = create_logger("Audit")
            logger.propagate = False
            self.audit_update = check_dataset(self.rootpath, self.datapath,
                                              outfile, filetype, tracking,  
                                              logger)
        else:
            if not outfile:
                outfile = self.outfile
            out = f'{out_dir}/{outfile}'
            if path.exists(out):
                remove(out)
            logger = create_logger_outfile(outfile, name = 'Audit')
            logger.propagate = False
            self.audit_update = check_dataset(self.rootpath, self.datapath,
                                              outfile, filetype, tracking,  
                                              logger)
            
        logger.info('')
        logger.info(self.audit_update)
        
        
    def list_dataset(self, outfile=False): 
        
        """Prints out all the files inside the directory <in_dir> along with the
        directories 1 and 2 levels above containing the files. You can specify 
        an optional output filename that will contain the printout.
        """

        if not outfile and not self.outfile:
            logger = create_logger('Audit')  
        else:
            if not outfile:
                outfile = self.outfile
            out_dir = f'{self.outpath}/audit'
            if not path.exists(out_dir):
                mkdir(out_dir)
            out = f'{out_dir}/{outfile}'
            if path.exists(out):
                remove(out)
            logger = create_logger_outfile(out, name='Audit')

        logger.propagate = False
        
        logger.info("")
        logger.info("")
        for dirPath, dirNames, fileNames in walk(self.datapath):
            try:
                fileNames.remove('.DS_Store')
            except(ValueError):
                pass
            
            if fileNames or dirPath.split('/')[-1]=='eeg':
                dir1 = dirPath.split('/')[-3]
                dir2 = dirPath.split('/')[-2]
                dir3 = dirPath.split('/')[-1]
                logger.info(f"Directory: {dir1}/{dir2}/{dir3}")
                logger.info(f"Files; {fileNames}")
                logger.info('-' * 10)

    
    def track(self, subs = 'all', ses = 'all', step = None, chan = None, 
                    stage = None, outfile = False, show = True, log = True):
        
        ## Set up logging
        logger = create_logger('Tracking')
        logger.info('')
        
        ## Set tracking variable
        if self.tracking:
            tracking = self.tracking
        else:
            tracking = {}
        
        ## Track sessions  
        if not isinstance(subs, list) and subs == 'all':
            try:
                subs = [x for x in listdir(self.datapath) if '.' not in x]
            except:
                logger.critical(f'{self.datapath} does not exist - cannot ascertain '
                             'details of dataset.')
                return
        elif not isinstance(subs, list):
            
            subs = read_tracking_sheet(self.rootpath, logger)
            subs = subs['sub'].drop_duplicates().tolist()
        subs.sort()
        
        # Tracking
        tracking['ses'] = {}
        for sub in subs:
            try:
                tracking['ses'][sub] = [x for x in listdir(f'{self.datapath}/{sub}') 
                                    if '.' not in x]
            except:
                logger.warning(f'No sessions found for {sub}')
                tracking['ses'][sub] = ['-']
            
        # Dataframe
        df = DataFrame(data=None, dtype=object)
        df.index = subs
        df['ses'] = '-'
        for x in df.index:
            df.loc[x,'ses'] = tracking['ses'][x]
        
        # Loop through other steps
        if step: 
            df, tracking = track_processing(self, step, subs, tracking, df, chan, 
                                                  stage, show, log)

        # Update tracking
        try:
            self.tracking = self.tracking | tracking
        except:
            self.tracking = {**self.tracking, **tracking}
        
        if show:
            logger.info('')
            logger.info(df)
        if outfile:
            df.to_csv(f'{self.outpath}/audit/{outfile}')

        return   

    def make_bids(self, subs = 'all', origin = 'SCN', filetype = '.edf',
                  outfile = True):
        # Set up logging
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/make_bids_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Make bids')
            logger.info('')
            logger.info(f"-------------- New call of 'Load sleep stages' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Make bids')
        else:
            logger = create_logger('Make bids')
        logger.info('')
        logger.debug('Formatting dataset into BIDS.')
        
        make_bids(self.datapath, subs, origin, filetype)
        
    def extract_channels(self, exclude = None):
        extract_channels(self.datapath, exclude)
        
    def QC_channels(self, subs = 'all', sessions = 'all', filetype = '.edf',
                    filt = None, chantype = ['eeg', 'eog', 'emg', 'ecg'], 
                    outfile=True):
        
        """
        Run automated quality control (QC) on physiological channels for all subjects and sessions.
    
        This method initializes a SQUID instance and performs QC on the specified channel types
        (e.g., EEG, EOG, EMG, ECG). For each channel, signal quality metrics are computed, such as:
    
        - `time_not_flatline`:        Percentage of time the signal is not flatlined.
        - `time_above_10`:            Percentage of signal above 10 µV.
        - `time_below_200`:           Percentage of signal below 200 µV.
        - `gini_coeff`:               Gini coefficient of signal inequality.
        - `inverse_power_ratio`:      Inverse power ratio of high-to-low frequency content.
        - `ecg_artefact`:             Mean correlation between EEG and ECG signals.
        - `ecg_artefact_perc`:        Percentage of signal with ECG artefact correlation r > 0.5.
    
        Parameters
        ----------
        subs : str or list, optional
            Subjects to include in QC. Can be 'all' or a list of BIDS-compatible subject IDs.
    
        sessions : str or list, optional
            Sessions to include in QC. Can be 'all' or a list of session labels.
    
        filetype : str, default='.edf'
            File format to look for in BIDS dataset (e.g., '.edf', '.set', '.vhdr').
            
        filt : list or None
            Explicit list of channels to process for EEG. If None, then standard 10-20 EEG channel 
            names will be used. 
    
        chantype : list of str, default=['eeg', 'eog', 'emg', 'ecg']
            Channel types to process. Accepted values: ['eeg', 'eog', 'emg', 'ecg'].
    
        outfile : bool or str, default=True
            If True, writes log to auto-generated timestamped file in `self.log_dir`.
            If str, writes log to the specified file path.
            If False, logs only to console.
    
        Returns
        -------
        None
            Results are stored in the SQUID object and/or exported downstream.
        """
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/load_sleep_stages_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile = logfile, name='Channel QC')
            logger.info('')
            logger.info(f"-------------- New call of 'Channel QC' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Channel QC')
        else:
            logger = create_logger('Channel QC')
        logger.info('')
        
        
        # Check chantypes
        valid_chantypes = {'eeg', 'eog', 'emg', 'ecg'}
        if not set(chantype).issubset(valid_chantypes):
            raise ValueError(f"Invalid chantype(s) specified. Allowed: {valid_chantypes}")
        
        # Run detection
        if subs == 'all':
            subs = None
        if sessions == 'all':
            sessions = None
        squid = SQUID(self.rootpath, filetype = filetype, subjects = subs, 
                     sessions = sessions )
        squid.process_all(filt, chantype)
        
    def QC_summary(self, qc_dir = None, chantype = ['eeg', 'eog', 'emg', 'ecg']):
        
        if not qc_dir:
            qc_root = self.outpath + '/QC'
            
        # output directory    
        out_dir = self.outpath + '/datasets/'
        if not path.exists(out_dir ):
            mkdir(out_dir)
        out_dir = self.outpath + '/datasets/QC'
        if not path.exists(out_dir):
            mkdir(out_dir)
        
        # Run and save
        for modality in chantype:
            summary = gather_qc_reports(qc_root, modality)
            summary.to_csv(f'{out_dir}/summary_{modality}.csv')
            
    
    def load_stages(self, xml_dir = None, subs = 'all', sessions = 'all', 
                          filetype = '.edf', stage_key = None, 
                          outfile = True):
        '''
            Extracts stages from the BIDS formatted dataset, in which
            staging has been listed in a file *acq-PSGScoring_events.tsv, and
            saves the information in an annotations (.xml) file
        '''
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/load_sleep_stages_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Load sleep stages')
            logger.info('')
            logger.info(f"-------------- New call of 'Load sleep stages' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Load sleep stages')
        else:
            logger = create_logger('Load sleep stages')
        logger.info('')
        
        # Set xml_dir
        if not xml_dir:
            xml_dir = self.outpath + '/staging_manual'
        if not path.exists(xml_dir):
            mkdir(xml_dir)
        
        # Load stages
        flag = load_stages(self.datapath, xml_dir, subs, sessions, filetype, 
                           stage_key)
        
        # Log finish 
        if flag > 0:
            logger.warning(f"'load_stages' finished with {flag} WARNINGS. See log for detail.")
        else:
            logger.debug("'load_stages' finished without error.")
    #--------------------------------------------------------------------------
    '''
    ANALYSIS FUNCTIONS
    
    power_spectrum -> performs power spectral analysis.
                    
    
    '''    
    
    
    def power_spectrum(self, xml_dir = None, out_dir = None, 
                             subs = 'all', sessions = 'all', filetype = '.edf',  
                             chan = None, ref_chan = None, 
                             grp_name = 'eeg', rater = None, 
                             stage = ['NREM1','NREM2','NREM3', 'REM'], 
                             cycle_idx = None, concat_cycle = True, 
                             concat_stage = False, general_opts = None, 
                             frequency_opts = None, filter_opts = None, 
                             epoch_opts = None, event_opts = None, 
                             norm = None, norm_opts = None, 
                             outfile = True):
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_power_spectrum_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Power spectrum')
            logger.info('')
            logger.info(f"-------------- New call of 'Power spectrum' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Power spectrum')
        else:
            logger = create_logger('Power spectrum')
        logger.info('')

        
        # Set input/output directories
        in_dir = self.datapath
        
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')

        if not out_dir:
            out_dir = f'{self.outpath}/powerspectrum' 
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {out_dir}')
        
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if isinstance(chan, str):
            return
        
        # Set default parameters
        if not general_opts:
            general_opts = default_general_opts()
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        if not norm_opts:
            norm_opts = default_norm_opts()
        if not filter_opts:
            filter_opts = default_filter_opts()    
        frequency_opts['frequency'] = (filter_opts['highpass'], filter_opts['lowpass'])
        
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),
               1,
               int(event_opts['concat_events']),
               )
        
        # Set suffix for output filename
        if not general_opts['suffix']:
            general_opts['suffix'] = f"{frequency_opts['frequency'][0]}-{frequency_opts['frequency'][1]}Hz"
        
        # Check annotations directory exists, run detection
        spectrum = Spectrum(in_dir, xml_dir, out_dir, chan, ref_chan, 
                            grp_name, stage, cat, rater, cycle_idx, subs, 
                            sessions, self.tracking)
                
        spectrum.powerspec_it(general_opts, frequency_opts, filter_opts, 
                              epoch_opts, event_opts, norm, norm_opts, 
                              filetype, logger) 
        
        try:
            self.tracking = self.tracking | spectrum.tracking
        except:
            self.tracking = {**self.tracking, **spectrum.tracking}
            
        return 
        

    #--------------------------------------------------------------------------
    '''
    SLEEP EVENTS DETECTIONS
    
    sleep_staging
    detect_artefacts,
    detect_spectral_peaks,
    detect_slow_oscillations,
    detect_spindles,
    
    
    '''
    def detect_sleep_stages(self, out_dir = None, 
                                  subs = 'all', sessions = 'all', filetype = '.edf', 
                                  method = 'Vallat2021', qual_thresh = 0.5,
                                  eeg_chan = None, ref_chan = None, 
                                  eog_chan = None, emg_chan = None, 
                                  rater = None, invert = False, outfile = True):
        
        from seapipe.events.seabass import seabass

        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/detect_sleep_stages_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect sleep stages')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect sleep stages' evoked at {today}, {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect sleep stages')
        else:
            logger = create_logger('Detect sleep stages')
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not out_dir:
            out_dir = f'{self.outpath}/staging_auto'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        track_sheet = read_tracking_sheet(self.rootpath, logger)
        if not subs:
            subs = [x for x in list(set(track_sheet['sub']))]
            subs.sort()
        if not sessions:
            sessions = track_sheet
        
        # Set channels
        eeg_chan, ref_chan = check_chans(self.rootpath, eeg_chan, ref_chan, logger)
        if isinstance(eeg_chan, str) and eeg_chan == 'error':
            return
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical("The argument 'invert' must be set to either: "
                            f"'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('')
            logger.info("Check documentation for how to set up staging data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            logger.critical('Sleep stage detection finished with ERRORS. See log for details.')
            return
    
        # Check annotations directory exists, run detection
        stages = seabass(in_dir, out_dir, eeg_chan, ref_chan, eog_chan, emg_chan, 
                         rater, subs, sessions, self.tracking) 
        stages.detect_stages(method, qual_thresh, invert, filetype, track_sheet,
                             logger)
        try:
            self.tracking = self.tracking | stages.tracking
        except:
            self.tracking = {**self.tracking, **stages.tracking}
        return
    
    
    def detect_artefacts(self, xml_dir = None, out_dir = None, 
                               subs = 'all', sessions = 'all', filetype = '.edf', 
                               method = 'seapipe', win_size = 5,
                               chan = None, ref_chan = None, 
                               label = 'individual',
                               rater = None, grp_name = 'eeg', 
                               stage = ['NREM1', 'NREM2', 'NREM3', 'REM'],
                               outfile = True):
        
        from seapipe.events.sand import SAND

        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_artefacts_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect artefacts')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect artefacts' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect artefacts')
        else:
            logger = create_logger('Detect artefacts')
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
            
        if not out_dir:
            out_dir = select_input_dirs(self.outpath, xml_dir, 'staging')   
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
        
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, 
                                          logger)
        
    
        # Check annotations directory exists, run detection
        artefacts = SAND(in_dir, xml_dir, out_dir, chan, ref_chan, 
                         rater, grp_name, subs, sessions, self.tracking) 
        artefacts.detect_artefacts(method, label,  win_size,  filetype, stage,
                                   logger)
                                   
    
        try:
            self.tracking = self.tracking | artefacts.tracking
        except:
            self.tracking = {**self.tracking, **artefacts.tracking}
        return
        
        
    
    def detect_spectral_peaks(self, xml_dir = None, out_dir = None, 
                                    subs = 'all', sessions = 'all', chan = None, 
                                    ref_chan = None, grp_name = 'eeg', 
                                    rater = None, frequency = (9,16), 
                                    stage = ['NREM2','NREM3'], cycle_idx = None,
                                    concat_cycle = True, concat_stage = False,
                                    general_opts = None, frequency_opts = None,
                                    filter_opts = None, epoch_opts = None, 
                                    event_opts = None, fooof_opts = None, 
                                    filetype = '.edf', suffix = None, 
                                    outfile = True):
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_specparams_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile = logfile, 
                                           name = 'Detect spectral peaks')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect spectral peaks' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile = logfile, 
                                           name = 'Detect spectral peaks')
        else:
            logger = create_logger('Detect spectral peaks')
        
        # Set input/output directories
        in_dir = self.datapath
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
            
        if not out_dir:
            out_dir = f'{self.outpath}/fooof' 
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {out_dir}')
            
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()    
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),1,1)
        
        # Run detection
        spectrum = Spectrum(in_dir, xml_dir, out_dir, chan, 
                            ref_chan, grp_name, stage, cat, rater, 
                            cycle_idx, subs, sessions, self.tracking)
        
        if not general_opts:
            general_opts = default_general_opts()
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not filter_opts:
            filter_opts = default_filter_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        if not fooof_opts:
            fooof_opts = default_fooof_opts() 
            
        fooof_opts['bands_fooof'] = [frequency]
        
        # Set suffix for output filename
        if not suffix:
            general_opts['suffix'] = f'{frequency[0]}-{frequency[1]}Hz'
        
        spectrum.fooof_it(general_opts, frequency_opts, filter_opts, 
                          epoch_opts, event_opts, fooof_opts, 
                          filetype, logger)  
                
        return 
    
    
    def detect_slow_oscillations(self, xml_dir=None, out_dir=None, subs='all', 
                                       sessions='all', filetype='.edf', 
                                       method = ['Staresina2015'], chan=None,
                                       ref_chan=None, rater=None, grp_name='eeg', 
                                       stage = ['NREM2','NREM3'], cycle_idx=None, 
                                       duration=(0.2, 2), invert = None,
                                       reject_artf = ['Artefact', 'Arou', 'Arousal'],
                                       average_channels = False, outfile = True):
        
        from seapipe.events.seasnakes import seasnakes

        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            evt_out = '_'.join(method)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_slowosc_{evt_out}_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect slow oscillations')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect slow oscillations' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect slow oscillations')
        else:
            logger = create_logger('Detect slow oscillations')
        
        logger.info('')
        logger.debug("Commencing SO detection pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
        
        if not out_dir:
            out_dir = f'{self.outpath}/slowwave'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical(f"The argument 'invert' must be set to either: 'True', "
                            f"'False' or 'None'; but it was set as {invert}.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('SO detection finished with ERRORS. See log for details.')
            return
            
        # Format concatenation
        cat = (1,1,1,1)
    
        # Run detection
        SO = seasnakes(in_dir, xml_dir, out_dir, chan, ref_chan, 
                         grp_name, stage, rater, subs, sessions, 
                         self.tracking, reject_artf) 
        SO.detect_slowosc(method, cat, cycle_idx, duration, 
                               average_channels, invert, filetype, logger)
        try:
            self.tracking = self.tracking | SO.tracking
        except:
            self.tracking = {**self.tracking, **SO.tracking}
        
        return
    
    
    def detect_spindles(self, xml_dir = None, out_dir = None, subs = 'all', 
                              sessions = 'all', filetype = '.edf', 
                              method = ['Moelle2011'], chan = None, 
                              ref_chan = None, rater = None, 
                              stage = ['NREM2','NREM3'], grp_name = 'eeg', 
                              cycle_idx = None, concat_cycle = True, 
                              frequency = None, adap_bands = 'Fixed', 
                              adap_bw = 4, duration = (0.5, 3),
                              reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                              outfile = True):
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            evt_out = '_'.join(method)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_spindles_{evt_out}_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect spindles' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles')
        else:
            logger = create_logger('Detect spindles')
            

        logger.info('')
        logger.debug("Commencing spindle detection pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath   
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
        
        if not out_dir:
            for met in method:
                out_dir = select_output_dirs(self.outpath, out_dir, met)  
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if not isinstance(chan, DataFrame) and not isinstance(chan, list):
            logger.error('Problem loading channel information')
            return
        elif isinstance(ref_chan, str):
            logger.error('Problem loading ref-channel information')
            return
        
        # Format concatenation
        if concat_cycle == True:
            cat = (1,0,1,1)
        else:
            cat = (0,0,1,1)
            if not cycle_idx:
                logger.error("'concat_cycle' is set to false, but 'cycle_idx' = None. "
                             "Set cycle_idx to a list of integers to use cycles properly.")
                logger.info("Check documentation for how to mark and use sleep cycles:")
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                return
        
        # Check for adapted bands
        logger.debug(f'Detection using {adap_bands} frequency bands has been selected.')
        frequency, flag = adap_bands_setup(self, adap_bands, frequency, subs, 
                                           sessions, chan, ref_chan, stage, cat, 
                                           concat_cycle, cycle_idx, logger)
        if flag == 'error':
            return
       
        # Run detection
        self.track(subs, sessions, step = ['fooof','spindle'], show = False, 
                   log = False)
        
        #self.track(step='fooof', show = False, log = False)
        #self.track(step='spindle', show = False, log = False)
        spindle = whales(self.rootpath, in_dir, xml_dir, out_dir, 
                         chan, ref_chan, grp_name, stage, frequency, rater, 
                         subs, sessions, reject_artf, self.tracking) 
        spindle.whale_it(method, cat, cycle_idx, adap_bands, adap_bw, 
                         duration, filetype, logger)
        try:
            self.tracking = self.tracking | spindle.tracking
        except:
            self.tracking = {**self.tracking, **spindle.tracking}
            
        return
    
    
    def whales(self, xml_dir = None, out_dir = None, subs = 'all', 
                     sessions = 'all', filetype = '.edf', 
                     method = ['Moelle2011', 'Ray2015'], evt_out = 'spindle',
                     merge_type = 'consensus', weights = None,
                     chan = None, ref_chan = None, rater = None, 
                     stage = ['NREM2','NREM3'], grp_name = 'eeg', 
                     cycle_idx = None, s_freq = None, keyword = None, 
                     duration =(0.5, 3),
                     reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                     outfile = True):
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_spindles_WHALES_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles (WHALES)')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect spindles (WHALES)' evoked at {now} --------------")
       
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles (WHALES)')
        else:
            logger = create_logger('Detect spindles (WHALES)')
            
        logger.info('')
        logger.debug("Commencing spindle optimisation pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        xml_dir = select_input_dirs(self.outpath, xml_dir, 'spindle') 
        logger.debug(f'Input annotations being read from: {xml_dir}')
        
        if not out_dir:
            out_dir = xml_dir  
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        if merge_type == 'consensus':
            cs_thresh = 0.5
        elif merge_type == 'addition':
            cs_thresh = 0.01
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
        
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if not isinstance(chan, DataFrame) and not isinstance(chan, list):
            logger.error('Problem loading channel information')
            return
        elif isinstance(ref_chan, str):
            logger.error('Problem loading ref-channel information')
            return
        
        self.track(step='spindle', show = False, log = False)
        
        logger.debug('Starting Merge Now')
        spindle = whales(self.rootpath, in_dir, xml_dir, out_dir, chan, ref_chan, 
                         grp_name, stage, frequency = None, rater = rater, 
                         subs = subs, sessions = sessions, 
                         reject_artf = reject_artf, tracking = self.tracking) 
        spindle.whales(method, merge_type, chan, rater, stage, ref_chan, grp_name, 
                       keyword, cs_thresh, s_freq, duration, evt_out, weights,
                       filetype, logger)
    
    
    def detect_rems(self, xml_dir = None, out_dir = None, subs = 'all', 
                          sessions = 'all', filetype = '.edf', 
                          method = ['YASA'], chan = None,
                          ref_chan = None, rater = None, grp_name = 'eeg', 
                          stage = ['REM'], cycle_idx = None, 
                          amplitude = (50, 325), duration = (0.3, 1.5),
                          reject_artf = ['Artefact', 'Arou', 'Arousal'],
                          average_channels = False, outfile = True):
        
        from seapipe.events.remora import remora

        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            evt_out = '_'.join(method)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{self.log_dir}/detect_rems_{evt_out}_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect eye movements (REMS)')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect rapid eye movements' evoked at {now} --------------")
            
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect Detect REMS')
        else:
            logger = create_logger('Detect Detect REMS')
        
        logger.info('')
        logger.debug("Commencing REMS detection pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
        
        if not out_dir:
            out_dir = f'{self.outpath}/rems'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
    
        # Run detection
        REMS = remora(in_dir, xml_dir, out_dir, chan, ref_chan, 
                      rater, grp_name, reject_artf, subs, sessions)
        
        REMS.detect_rems(method, amplitude, duration, stage, cycle_idx, 
                         filetype, logger)
        try:
            self.tracking = self.tracking | REMS.tracking
        except:
            self.tracking = {**self.tracking, **REMS.tracking}
            
        return
    
    #--------------------------------------------------------------------------
    '''
    PLOTTING.
    
    event_spectrogram ->
    
    
    '''
    
    
    def spectrogram(self, xml_dir = None, out_dir = None, subs = 'all', 
                          sessions = 'all', filetype = '.edf', chan = None, 
                          ref_chan = None, rater = None, stage = None, 
                          grp_name = 'eeg', cycle_idx = None, 
                          concat_stage = False, concat_cycle = True, 
                          evt_type = None, buffer = 0, invert = None, 
                          filter_opts = None, progress=True, outfile=False):
        
        from seapipe.spectrum.spectrogram import event_spectrogram

        # Set up logging
        logger = create_logger('Event spectrogram')
        logger.info('')
        logger.debug("Creating spectrogram of events.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
            
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'staging') 
            
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            "run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
        
        if not out_dir:
            out_dir = f'{self.outpath}/spindle'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),1,1)
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, chan, False, logger)
        elif type(invert) != bool:
            logger.critical("The argument 'invert' must be set to either: "
                            f"'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        if not filter_opts:
            filter_opts = default_filter_opts()
        
        
        if not evt_type:
            logger.warning("No event type (evt_type) has been specified. "
                           "Spectrogram will be run on epochs instead. "
                           "This may take some time...")
        
        event_spectrogram(self, in_dir, xml_dir, out_dir, subs, sessions, stage, 
                                cycle_idx, chan, ref_chan, rater, grp_name, 
                                evt_type, buffer, invert, cat, filter_opts,  
                                outfile, filetype, progress, self.tracking)
        
        return
    
    
    #--------------------------------------------------------------------------
    '''
    PHASE AMPLITUDE COUPLING.
    
    mean_amps -> runs Phase Amplitude Coupling analyses on sleep EEG data. 
    
    
    '''    
    def pac(self, xml_dir = None, out_dir = None, subs = 'all', sessions = 'all', 
                  filetype = '.edf', chan = None, ref_chan = None, rater = None, 
                  grp_name = 'eeg', stage = ['NREM2','NREM3'], concat_stage = True, 
                  cycle_idx = None, concat_cycle = True,  
                  method = 'MI', surrogate = 'Time lag', correction = 'Z-score',
                  adap_bands_phase = 'Fixed', frequency_phase = (0.5, 1.25), 
                  adap_bands_amplitude = 'Fixed', frequency_amplitude = (11, 16),
                  adap_bw = 4, min_dur = 1, nbins = 18, invert = None,
                  frequency_opts = None, filter_opts = None, epoch_opts = None, 
                  evt_name = None, event_opts = None, 
                  reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                  progress = True, outfile = True):
        
        from seapipe.pac.octopus import octopus, pac_method
        from seapipe.pac.pacats import pacats

        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            pha = f'{frequency_phase[0]}-{frequency_phase[1]}'
            amp = f'{frequency_amplitude[0]}-{frequency_amplitude[1]}'
            logfile = f'{self.log_dir}/pac_{pha}_{amp}_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Phase-amplitude coupling')
            logger.info('')
            logger.info(f"-------------- New call of 'Phase-amplitude coupling' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Phase-amplitude coupling')
        else:
            logger = create_logger('Phase-amplitude coupling')
        logger = create_logger('Phase-amplitude coupling')
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name) 

        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been "
                            f"run or Events: {evt_name} haven't been detected.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        else:    
            logger.debug(f'Input annotations being read from: {xml_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if not isinstance(chan, DataFrame) and not isinstance(chan, list):
            return
        elif isinstance(ref_chan, str):
            return
        
        # Check for adapted bands 
        cat = (int(concat_cycle),int(concat_stage),0,0)
        
        frequency_phase, flag1 = adap_bands_setup(self, adap_bands_phase, frequency_phase, 
                                     subs, sessions, chan, ref_chan, stage, cat, 
                                     concat_cycle, cycle_idx, logger)
        
        frequency_amplitude, flag2 = adap_bands_setup(self, adap_bands_amplitude, 
                                               frequency_amplitude, subs, sessions, 
                                               chan, ref_chan, stage, cat, 
                                               concat_cycle, cycle_idx, logger)
        
        if flag1 == 'error' or flag2 == 'error':
            return
        
        # Set PAC methods
        idpac = pac_method(method, surrogate, correction)
        
        # Set default parameters
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        if not filter_opts:
            filter_opts = default_filter_opts()   
        filter_opts['bandpass'] = False
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical("The argument 'invert' must be set to either: "
                            f"'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Phase amplitude coupling finished with ERRORS. See log for details.')
            return
        
        self.track(subs, sessions, step = 'fooof', show = False, 
                   log = False)
        self.tracking['event_pac'] = {}
        
        # Check whether event based or continuous
        if evt_name: #OCTOPUS
            if not out_dir:
                out_dir = f'{self.outpath}/event_pac'  
            if not path.exists(out_dir):
                mkdir(out_dir)
            logger.debug(f'Output being saved to: {out_dir}')
                
            cat = (int(concat_cycle),int(concat_stage),0,0)
            Octopus = octopus(self.rootpath, in_dir, xml_dir, out_dir, 
                              chan, ref_chan, grp_name, stage, rater, 
                              subs, sessions, reject_artf,
                              self.tracking)
            
            Octopus.pac_it(cycle_idx, cat, nbins, filter_opts, epoch_opts, 
                           frequency_opts, event_opts, filetype, idpac, evt_name, 
                           min_dur, adap_bands_phase, frequency_phase, 
                           adap_bands_amplitude, frequency_amplitude, 
                           adap_bw, invert, progress, logger)
        else: #PACATS
            if not out_dir:
                out_dir = f'{self.outpath}/pac'  
            if not path.exists(out_dir):
                mkdir(out_dir)
            logger.debug(f'Output being saved to: {out_dir}')
            cat = (int(concat_cycle),int(concat_stage),1,1)
            Pacats = pacats(self.rootpath, in_dir, xml_dir, out_dir, 
                            chan, ref_chan, grp_name, stage, rater, subs, sessions, 
                            reject_artf, self.tracking)
            Pacats.pac_it(cycle_idx, cat, nbins, filter_opts, epoch_opts, 
                           frequency_opts, filetype, idpac, 
                           min_dur, adap_bands_phase, frequency_phase, 
                           adap_bands_amplitude, frequency_amplitude,
                           adap_bw, invert, progress, logger)

        return
    
    '''
    DYNAMICS
    
    spindle_clustering -> analysis of spindle event temporal clustering 
    
    sigma_fluctuations -> 
    '''
    
    def cluster_flucs(self, evt_name, xml_dir = None, out_dir = None,
                            freq_bands = {'SWA': (0.5, 4), 'Sigma': (10, 15)},
                            subs = 'all', sessions = 'all', 
                            filetype = '.edf',
                            chan = None, ref_chan = None, grp_name = 'eeg', 
                            stage = ['NREM2'], 
                            rater = None, 
                            outfile = True):
        
        from seapipe.events.clam import clam

        # Force evt_name into list, and loop through events    
        if isinstance(evt_name, str):
            evts = [evt_name]
        elif isinstance(evt_name, list):
            evts = evt_name
        else:
            raise TypeError(f"'evt_name' can only be a str or a list, but {type(evt_name)} was passed.")
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            evt_out = '_'.join(evt_name)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/event_clustering_{evt_out}_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Event clustering')
            logger.info('')
            logger.debug(f"-------------- New call of 'Event clustering evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Event clustering')
        else:
            logger = create_logger('Event clustering')
        
        for evt_name in evts:
            
            out_dir = select_output_dirs(self.outpath, out_dir, 'cluster')
            logger.debug(f'Output being save to: {out_dir}')
            
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'cluster')
            logger.debug(f'Input annotations being read from: {xml_dir}')
            
            # Check annotations directory exists
            if not path.exists(xml_dir):
                logger.info('')
                logger.critical(f"{xml_dir} doesn't exist. Event detection has not "
                                "been run or an incorrect event type has been selected.")
                logger.info('Check documentation for how to run a pipeline:')
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                return
    
            # Set channels
            chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
            
            self.track(subs, sessions, step = 'spindle', show = False, 
                       log = False)
            self.tracking['cluster'] = self.tracking['spindle']
            
            # Run analysis
            CLAM = clam(self.rootpath, self.datapath, xml_dir, out_dir, 
                        chan, grp_name, stage, 
                        rater, subs, sessions, self.tracking)
            
            CLAM.clustering(evt_name, freq_bands, filetype, grp_name, logger )
        
        return
    
    
    #--------------------------------------------------------------------------
    '''
    DATASET CREATION.
    
    export_macro_stats -> Exports sleep macroarchitecture per participant into 
                            the corresponding folder in output directory 'staging' 
    
    macro_dataset -> Creates a cohort dataset of sleep macroarchitecture and saves
                        it to a single .csv file in output directory 'dataset'
    
    export_eventparams -> Exports descriptives for sleep events per participant into 
                            the corresponding folder in output directory 'staging'
    
    event_dataset -> Creates a cohort dataset of sleep events descriptives and saves
                        it to a single .csv file in output directory 'dataset'
    
    '''    
    
    def export_macro_stats(self, xml_dir = None, out_dir = None, 
                                 subs = 'all', sessions = 'all', 
                                 times = None, rater = None, 
                                 arousal_name = ['Arousal', 'Arou'], 
                                 outfile = True):
        
        # Set up logging
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/export_sleep_macro_stats_{today}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
            logger.info('')
            logger.info(f"-------------- New call of 'Export macro stats' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
        else:
            logger = create_logger('Export macro stats')

        
        # Set input/output directories
        xml_dir = select_input_dirs(self.outpath, xml_dir, 'macro')
        logger.debug(f'Input annotations being read from: {xml_dir}')
        
        out_dir = select_output_dirs(self.outpath, out_dir, 'macro')
        logger.debug(f'Output being saved to: {out_dir}')
        
        # Set channels
        times, ref_chan = check_chans(self.rootpath, None, True, logger)
        
        self.track(subs = subs, ses = sessions, step = ['staging'], show = False, 
                   log = True)
        
        sleepstats.export_sleepstats(xml_dir, out_dir, subs, sessions, 
                                     rater, times, arousal_name, logger)
        return
    
    def macro_dataset(self, xml_dir = None, out_dir = None, 
                      subs = 'all', sessions = 'all', outfile = True):
         
         # Set input/output directories
         if outfile == True:
             today = date.today().strftime("%Y%m%d")
             now = datetime.now().strftime("%H:%M:%S")
             logfile = f'{self.log_dir}/export_sleep_macro_stats_{today}_log.txt'
             logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
             logger.info('')
             logger.info(f"-------------- New call of 'Macro dataset' evoked at {now} --------------")
         elif outfile:
             logfile = f'{self.log_dir}/{outfile}'
             logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
         else:
             logger = create_logger('Export macro stats')
         if not path.exists(self.outpath + '/datasets/'):
             mkdir(self.outpath + '/datasets/')
         out_dir = self.outpath + '/datasets/macro/'
         
         xml_dir = select_input_dirs(self.outpath, xml_dir, 'macro')
         logger.debug(f'Input annotations being read from: {xml_dir}')
         
         out_dir = select_output_dirs(self.outpath, out_dir, 'macro')
         logger.debug(f'Output being save to: {out_dir}')
         
         sleepstats.sleepstats_from_csvs(xml_dir, out_dir,   
                                 subs, sessions, logger)
         return
    
    def export_eventparams(self, evt_name, frequency = None,
                                 xml_dir = None, out_dir = None, subs = 'all', 
                                 sessions = 'all', filetype = '.edf',
                                 chan = None, ref_chan = None, 
                                 stage = ['NREM2','NREM3'], grp_name = 'eeg', 
                                 rater = None, cycle_idx = None, 
                                 concat_cycle = True, concat_stage = False, 
                                 keyword = None, segs = None,  
                                 adap_bands = 'Fixed',  
                                 adap_bw = 4, params = 'all', epoch_dur = 30, 
                                 average_channels = False, outfile = True):
        
        # Force evt_name into list, and loop through events    
        if isinstance(evt_name, str):
            evts = [evt_name]
        elif isinstance(evt_name, list):
            evts = evt_name
        else:
            raise TypeError(f"'evt_name' can only be a str or a list, but {type(evt_name)} was passed.")
        
        # Set up logging
        if outfile == True:
            subs_str, ses_str = out_names(subs, sessions)
            evt_out = '_'.join(evt_name)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/export_params_{evt_out}_subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Export params')
            logger.info('')
            logger.debug(f"-------------- New call of 'Export params' evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Export params')
        else:
            logger = create_logger('Export params')
            
            
        frequency, flag = adap_bands_setup(self, adap_bands, None, None, 
                                           None, None, None, stage, None, 
                                           concat_cycle, cycle_idx, logger)
        if flag == 'error':
            return
        
        # Set input/output directories
        in_dir = self.datapath
        
        for evt_name in evts:
            
            out_dir = select_output_dirs(self.outpath, out_dir, evt_name)
            logger.debug(f'Output being save to: {out_dir}')
            
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name)
            logger.debug(f'Input annotations being read from: {xml_dir}')
            
            # Check annotations directory exists
            if not path.exists(xml_dir):
                logger.info('')
                logger.critical(f"{xml_dir} doesn't exist. Event detection has not "
                                "been run or an incorrect event type has been selected.")
                logger.info('Check documentation for how to run a pipeline:')
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                return
    
            if adap_bands in ['Auto','Manual']:
                evt_name = f'{evt_name}_adap'
                self.track(step='fooof', show = False, log = False)
                peaks = check_chans(self.rootpath, None, False, logger)
            else:
                peaks = None
            
            # Set channels
            chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
            if average_channels:
                Ngo = {'run':True}
            else:
                Ngo = {'run':False}
            
            # Format concatenation
            cat = (int(concat_cycle),int(concat_stage),1,1)
            
            # Run line
            fish = FISH(self.rootpath, in_dir, xml_dir, out_dir, chan, ref_chan, grp_name, 
                              stage, rater, subs, sessions, self.tracking) 
            fish.line(keyword, evt_name, cat, segs, cycle_idx, frequency, adap_bands, 
                      peaks, adap_bw, params, epoch_dur, Ngo, filetype, logger)
        return
    
    
    def event_dataset(self, chan, evt_name, xml_dir = None, out_dir = None, 
                            subs = 'all', sessions = 'all', 
                            stage = ['NREM2','NREM3'], concat_stage = False, 
                            concat_cycle = True, cycle_idx = None, 
                            grp_name = 'eeg', adap_bands = 'Fixed',  
                            params = 'all', outfile = True):
        
        # Set up logging
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/event_dataset_{evt_name}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Event dataset')
            logger.info('')
            logger.info(f'-------------- New call of Event dataset evoked at {now} --------------')
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Event dataset')
        else:
            logger = create_logger('Event dataset')
        logger = create_logger('Event dataset')
        
        # Force evt_name into list, and loop through events    
        if isinstance(evt_name, str):
            evts = [evt_name]
        elif isinstance(evt_name, list):
            evts = evt_name
        else:
            logger.error(TypeError("'evt_name' can only be a str or a list of str, "
                                   f"but {type(evt_name)} was passed."))
            logger.info('Check documentation for how to create an event_dataset:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        frequency, flag = adap_bands_setup(self, adap_bands, None, None, 
                                           None, None, None, stage, None, 
                                           concat_cycle, cycle_idx, logger)
        if flag == 'error':
            return
        
        for evt_name in evts:
            # Append 'adap' after event name if adapted bands were used
            if adap_bands in ['Auto', 'Manual']:
                evt_name = f'{evt_name}_adap'
                self.track(step='fooof', show = False, log = False)
            
            # Set input/output directories
            in_dir = self.datapath
            if not out_dir:    
                if not path.exists(self.outpath + '/datasets/'):
                    mkdir(self.outpath + '/datasets/')
                outpath = self.outpath + f'/datasets/{evt_name}'
            else:
                outpath = out_dir
            if not path.exists(outpath):
                mkdir(outpath)
            logger.debug(f'Output being saved to: {outpath}')
            
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name)
            logger.debug(f'Input annotations being read from: {xml_dir}')
            if not path.exists(xml_dir):
                logger.info('')
                logger.critical(f"{xml_dir} doesn't exist. Event detection has not "
                                "been run or an incorrect event type has been selected.")
                logger.info('Check documentation for how to run a pipeline:')
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                return
            
            # Format concatenation
            cat = (int(concat_cycle),int(concat_stage),1,1)
            
            # Format chan
            if isinstance(chan, str):
                chan = [chan]
            
        
            fish = FISH(self.rootpath, in_dir, xml_dir, outpath, chan, None, grp_name, 
                              stage, subs = subs, sessions = sessions) 
            fish.net(chan, evt_name, adap_bands, params,  cat, cycle_idx, logger)
        
        return
    
    
    def pac_dataset(self, chan, evt_name = None, subs = 'all', sessions = 'all',
                          xml_dir = None, out_dir = None,  stage = None, 
                          concat_stage = False, concat_cycle = True, 
                          cycle_idx = None, grp_name = 'eeg', 
                          adap_bands_phase = 'Fixed', frequency_phase = (0.5, 1.25), 
                          adap_bands_amplitude = 'Fixed', frequency_amplitude = (11, 16),  
                          params = 'all', outfile=True):
        
        # Set up logging
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/PAC_dataset_{evt_name}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='PAC dataset')
            logger.info('')
            logger.info(f'-------------- New call of PAC dataset evoked at {now} --------------')
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='PAC dataset')
        else:
            logger = create_logger('PAC dataset')
        logger = create_logger('PAC dataset')
        
        # Set input/output directories
        in_dir = self.datapath

        if not out_dir:
            if not path.exists(self.outpath + '/datasets/'):
                mkdir(self.outpath + '/datasets/')
            out_dir = (f'{self.outpath}/datasets/event_pac' if evt_name 
                       else f'{self.outpath}/datasets/pac') 
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {out_dir}')
        
        # Check if event-based or continuous PAC
        if isinstance(evt_name, str):
            evt = 'event_pac' 
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt)
            cat = (int(concat_cycle),int(concat_stage),1,1)
        elif evt_name == None:
            evt = 'pac' 
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt)
            cat = (int(concat_cycle),int(concat_stage),0,0)
        else:
            logger.error(TypeError(f"'evt_name' can only be a str or NoneType, but {type(evt_name)} was passed."))
            logger.info('Check documentation for how to create a PAC summary dataset:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. PAC detection has not been "
                            "run or an incorrect type has been selected.")
            logger.info('Check documentation for how to run a pipeline:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return

        # Format chan
        if isinstance(chan, str):
            chan = [chan]
        
        # Default stage
        if stage == None:
            stage = ['NREM2','NREM3']
        
        # Run extraction
        self.track(subs, sessions, step = ['fooof','spindle'], show = False, 
                   log = False)    
        
        fish = FISH(self.rootpath, in_dir, xml_dir, out_dir, chan, None, grp_name, 
                          stage, None, subs, sessions, self.tracking) 
                   
        fish.pac_summary(chan, evt_name, adap_bands_phase, frequency_phase, 
                              adap_bands_amplitude, frequency_amplitude,
                              params = 'all', cat = cat, cycle_idx = None, 
                              logger = logger)
        
        return
    
    
    def cluster_dataset(self, chan, evt_name, xml_dir = None, out_dir = None, 
                                subs = 'all', sessions = 'all', 
                                stage = ['NREM2'], freq_bands = ('SWA', 'Sigma'),
                                params = 'all', outfile=True):
        
        # Set up logging
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/cluster_dataset_{evt_name}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Event dataset')
            logger.info('')
            logger.info(f'-------------- New call of Cluster dataset evoked at {now} --------------')
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Event dataset')
        else:
            logger = create_logger('Event dataset')
        logger = create_logger('Event dataset')
        
        # Force evt_name into list, and loop through events    
        if isinstance(evt_name, str):
            evts = [evt_name]
        elif isinstance(evt_name, list):
            evts = evt_name
        else:
            logger.error(TypeError("'evt_name' can only be a str or a list of str, "
                                   f"but {type(evt_name)} was passed."))
            logger.info('Check documentation for how to create an event_dataset:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        for evt_name in evts:    
            # Set input/output directories
            in_dir = self.datapath
            if not out_dir:    
                if not path.exists(self.outpath + '/datasets/'):
                    mkdir(self.outpath + '/datasets/')
                outpath = self.outpath + '/datasets/cluster_fluc'
                if not path.exists(outpath):
                    mkdir(outpath)
                outpath = outpath + f'/{evt_name}'
            else:
                outpath = out_dir
            if not path.exists(outpath):
                mkdir(outpath)
            logger.debug(f'Output being saved to: {outpath}')
            
            xml_dir = select_input_dirs(self.outpath, xml_dir, 'cluster')
            logger.debug(f'Input annotations being read from: {xml_dir}')
            if not path.exists(xml_dir):
                logger.info('')
                logger.critical(f"{xml_dir} doesn't exist. Cluster detection has not "
                                "been run.")
                logger.info('Check documentation for how to run a pipeline:')
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                return
            
            # Format chan
            if isinstance(chan, str):
                chan = [chan]
            
            fish = FISH(self.rootpath, in_dir, xml_dir, outpath, chan, None, 'eeg', 
                        stage, subs, sessions) 
            
            fish.scalops(chan, evt_name, freq_bands, params, logger)
        
        return
    

    def powerspec_dataset(self, chan, xml_dir = None, out_dir = None, 
                                subs = 'all', sessions = 'all', 
                                stage = ['NREM1','NREM2','NREM3', 'REM'], 
                                concat_stage = False, concat_cycle = True, 
                                cycle_idx = None, grp_name = 'eeg', 
                                rater = None, params = 'all', 
                                general_opts = None, frequency_opts = None, 
                                filter_opts = None, epoch_opts = None, 
                                event_opts = None, outfile=True):
        
        # Set up logging
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/powerspec_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Power spectrum dataset')
            logger.info('')
            logger.info(f'-------------- New call of Power spectrum dataset evoked at {now} --------------')
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Power spectrum dataset')
        else:
            logger = create_logger('Power spectrum dataset')
        logger = create_logger('Power spectrum dataset')
        
        # Set input/output directories
        in_dir = self.datapath
        log_dir = self.outpath + '/audit/logs/'
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not out_dir:
            if not path.exists(self.outpath + '/datasets/'):
                mkdir(self.outpath + '/datasets/')
            out_dir = f'{self.outpath}/datasets/powerspectrum'
            if not path.exists(out_dir):
                mkdir(out_dir)
        logger.debug(f'Output being saved to: {out_dir}')
        
        xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name = 'powerspectrum')
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Event detection has not "
                            "been run or an incorrect event type has been selected.")
            logger.info('Check documentation for how to run a pipeline:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        # Set default parameters
        if not general_opts:
            general_opts = default_general_opts()
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        
        if not filter_opts:
            filter_opts = default_filter_opts()    
        frequency_opts['frequency'] = (filter_opts['highpass'], filter_opts['lowpass'])
        
        # Set suffix for output filename
        if not general_opts['suffix']:
            general_opts['suffix'] = f"{frequency_opts['frequency'][0]}-{frequency_opts['frequency'][1]}Hz"
        
        # Format chan
        if isinstance(chan,str):
            chan = [chan]
        
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),1,1)
                            
        spectrum = Spectrum(in_dir, xml_dir, out_dir, chan, None, grp_name, 
                            stage,cat, rater, cycle_idx, subs, sessions) 
        spectrum.powerspec_summary(chan, general_opts, frequency_opts, filter_opts, 
                                   epoch_opts, event_opts, logger)
        
        return

    def bandpower_timecourse(self, band, xml_dir = None, out_dir = None,
                                   subs = 'all', sessions = 'all',
                                   filetype = '.edf', chan = None,
                                   ref_chan = None, grp_name = 'eeg',
                                   rater = None,
                                   stage = ['NREM1','NREM2','NREM3', 'REM'],
                                   cycle_idx = None, concat_cycle = True,
                                   concat_stage = True,
                                   general_opts = None, filter_opts = None,
                                   epoch_opts = None, event_opts = None,
                                   bandpower_opts = None, outfile = True):
        
        """Convenience wrapper that exports band-limited power vs. time."""

        # Local import avoids touching module-level imports per request
        from seapipe.spectrum.bandpower import (
            BandPowerTimecourse,
            default_bandpower_opts,
        )

        # Logging
        if outfile is True:
            subs_str, ses_str = out_names(subs, sessions)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = (f'{self.log_dir}/bandpower_timecourse_'
                       f'subs-{subs_str}_ses-{ses_str}_{today}_{now}_log.txt')
            logger = create_logger_outfile(logfile=logfile,
                                           name='Bandpower timecourse')
            logger.info('')
            logger.info("-------------- New call of 'Bandpower timecourse' "
                        f"evoked at {now} --------------")
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile,
                                           name='Bandpower timecourse')
        else:
            logger = create_logger('Bandpower timecourse')
        logger.info('')

        # Directories
        in_dir = self.datapath
        xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name = 'staging')
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not "
                            "been run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return

        if not out_dir:
            out_dir = f'{self.outpath}/bandpower_timecourse'
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {out_dir}')

        # Channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if isinstance(chan, str):
            return

        # Defaults
        if general_opts is None:
            general_opts = default_general_opts()
        if filter_opts is None:
            filter_opts = default_filter_opts()
        if epoch_opts is None:
            epoch_opts = default_epoch_opts()
        if event_opts is None:
            event_opts = default_event_opts()
        if bandpower_opts is None:
            bandpower_opts = default_bandpower_opts()

        # Concatenation mask for Spectrum
        cat = (int(concat_cycle),
               int(concat_stage),
               1,
               int(event_opts['concat_events']))

        tracker = BandPowerTimecourse(in_dir, xml_dir, out_dir, chan, ref_chan,
                                      grp_name, stage, cat, rater, cycle_idx,
                                      subs, sessions, self.tracking)
        tracker.bandpower_timecourse(
            band=band,
            general_opts=general_opts,
            filter_opts=filter_opts,
            epoch_opts=epoch_opts,
            bandpower_opts=bandpower_opts,
            event_opts=event_opts,
            filetype=filetype,
            split_by_stage=not concat_stage,
            logger=logger,
        )

        return


def out_names(subs, sessions):
    if isinstance(subs, list):
        subs_str = "_".join(subs).replace('\n', '').replace('\r', '').replace('sub','')
    else:
        subs_str = subs
    if isinstance(sessions, list):
        ses_str = "_".join(sessions).replace('\n', '').replace('\r', '').replace('ses','')
    else:
        ses_str = sessions     
        
    return subs_str, ses_str
        
